__all__ = ["interp_2d"]

import numpy as np
import scipy.signal as sps
from scipy.interpolate import interp2d
from scipy.ndimage import gaussian_filter


def interp_2d(data, ratio=10):
    # scipy interp. cubic
    X = np.arange(0, data.shape[1], 1)
    Y = np.arange(0, data.shape[0], 1)
    f = interp2d(X, Y, data, kind="cubic")
    xnew = np.arange(0, data.shape[1], 1.0 / ratio)
    ynew = np.arange(0, data.shape[0], 1.0 / ratio)
    data1 = f(xnew, ynew)
    # data1 = gaussian_filter(data1, sigma=3)
    Xn, Yn = np.meshgrid(xnew, ynew)

    return Xn[:-ratio, :-ratio], Yn[:-ratio, :-ratio], data1[:-ratio, :-ratio]



__doc__ = """

.. autoclass:: miv.core.pipeline.Pipeline
   :members:

"""
__all__ = ["Pipeline"]

from typing import List, Optional, Union

import pathlib

from miv.core.operator.chainable import _Chainable
from miv.core.policy import _Runnable


class Pipeline:
    """
    A pipeline is a collection of operators that are executed in a specific order.

    If operator structure is a tree like

    .. mermaid::

        flowchart LR
            A --> B --> D --> F
            A --> C --> E --> F
            B --> E

    then the execution order of `Pipeline(F)` is A->B->D->C->E->F.
    If nodes already have cached results, they will be loaded instead of being executed.
    For example, if E is already cached, then the execution order of `Pipeline(F)` is A->B->D->F. (C is skipped, E is loaded from cache)
    """

    def __init__(self, node: _Chainable):
        self.execution_order: List[_Runnable] = node.topological_sort()

    def run(
        self,
        working_directory: Optional[Union[str, pathlib.Path]] = "./results",
        no_cache: bool = False,
        skip_plot: bool = False,
        dry_run: bool = False,
        verbose: bool = False,  # Use logging
    ):
        """
        Run the pipeline.

        Parameters
        ----------
        working_directory : Optional[Union[str, pathlib.Path]], optional
            The working directory where the pipeline will be executed. By default "./results"
        no_cache : bool, optional
            If True, the cache will be disabled. By default False
        dry_run : bool, optional
            If True, the pipeline will not be executed. By default False
        verbose : bool, optional
            If True, the pipeline will log debugging informations. By default False
        """
        for node in self.execution_order:
            if verbose:
                print("Running: ", node)
            if hasattr(node, "cacher"):
                node.cacher.cache_policy = "OFF" if no_cache else "AUTO"
            node.run(dry_run=dry_run, save_path=working_directory, skip_plot=skip_plot)
        if verbose:
            print("Pipeline done:")
            self.summarize()
            print("-" * 46)

    def summarize(self):
        strs = []
        strs.append("Execution order:")
        for i, op in enumerate(self.execution_order):
            strs.append(f"{i}: {op}")
        return "\n".join(strs)

    def export(self, working_directory: Optional[Union[str, pathlib.Path]]):
        # TODO
        pass



__doc__ = """

Useful wrapper functions for MIV operators.

.. autofunction:: miv.core.wrapper.wrap_cacher

.. autofunction:: miv.core.wrapper.wrap_generator_to_generator

"""

__all__ = [
    "wrap_cacher",
    "wrap_generator_to_generator",
]

import types
from typing import Protocol, Union

import functools
import inspect
from collections import UserList
from dataclasses import dataclass, make_dataclass

from miv.core.datatype import DataTypes, Extendable
from miv.core.operator.operator import Operator, OperatorMixin


def wrap_cacher(cache_tag):
    """
    Decorator to wrap the function to use cacher.
    """

    def decorator(func):
        # @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not hasattr(wrapper, "cache_tag"):
                setattr(wrapper, "cache_tag", cache_tag)
            self: Operator = args[0]
            self.cacher.cache_tag = wrapper.cache_tag
            if self.cacher.check_cached():
                return next(self.cacher.load_cached())
            else:
                result = func(*args, **kwargs)
                self.cacher.save_cache(result)
                self.cacher.save_config()
                return result

        return wrapper

    return decorator


def wrap_generator_to_generator(func):
    """
    If all input arguments are generator, then the output will be a generator.

    TODO: Fix to accomodate free functions
    Current implementation only works for class methods.
    """

    # @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        is_all_generator = all(inspect.isgenerator(v) for v in args[1:]) and all(
            inspect.isgenerator(v) for v in kwargs.values()
        )
        if is_all_generator:
            if self.cacher.check_cached():

                def generator_func(*args, **kwargs):
                    yield from self.cacher.load_cached()

            else:

                def generator_func(*args, **kwargs):
                    for idx, zip_arg in enumerate(zip(*args)):
                        result = func(self, *zip_arg, **kwargs)
                        self.cacher.save_cache(result, idx)
                        yield result
                    else:
                        self.cacher.save_config()

            return generator_func(*args, **kwargs)
        else:
            if self.cacher.check_cached():
                return next(self.cacher.load_cached())
            else:
                result = func(self, *args, **kwargs)
                self.cacher.save_cache(result)
                self.cacher.save_config()
            return result

    return wrapper


def miv_function(name, **params):
    """
    Decorator to convert a function into a MIV operator.
    """

    def decorator(func):
        _LambdaClass = make_dataclass(
            name,
            [k for k in params.keys()],
        )

        @dataclass
        class LambdaClass(_LambdaClass, OperatorMixin):
            tag: str = name

            def __call__(self, *args, **kwargs):
                return func(self, *args)

            def __post_init__(self):
                _LambdaClass.__init__(self, **params)
                OperatorMixin.__init__(self)

        LambdaClass.__name__ = name
        obj = LambdaClass(**params)
        return obj

    return decorator


def test_wrap_miv_function():
    @miv_function("testTag", a=1, b=2)
    def func(self, c):
        print("here")
        print(c)

    print(func)
    print(func.__dir__())
    print(func(1))


if __name__ == "__main__":
    test_wrap_miv_function()



__all__ = [
    "firing_rates",
    "MFRComparison",
    "interspike_intervals",
    "coefficient_variation",
    "binned_spiketrain",  # Deprecated: Remove in the future version ^0.3.0
    "fano_factor",
    "decay_spike_counts",
    "spike_counts_with_kernel",
]

from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import datetime
import os
from dataclasses import dataclass

import elephant.statistics
import matplotlib.pyplot as plt
import neo
import numpy as np
import quantities as pq
import scipy
import scipy.signal

from miv.core.datatype import Spikestamps
from miv.core.operator import OperatorMixin
from miv.typing import SpikestampsType


def firing_rates(spiketrains: Spikestamps) -> Dict[str, Any]:
    """
    Process basic spiketrains statistics: rates, mean, variance.

    Parameters
    ----------
    spiketrains : Iterable[neo.core.SpikeTrain]

    Returns
    -------
    Iterable[Any]

    """
    rates = []
    if sum(spiketrains.get_count()) == 0:
        return {
            "rates": [0.0 for _ in range(spiketrains.number_of_channels)],
            "mean": 0.0,
            "variance": 0.0,
        }
    for spikestamp in spiketrains.neo():
        if len(spikestamp) == 0:
            rates.append(0)
            continue
        mfr = elephant.statistics.mean_firing_rate(spikestamp)
        if isinstance(mfr, pq.quantity.Quantity):
            mfr = mfr.magnitude[()]
        rates.append(mfr)

    rates_mean_over_channel = np.mean(rates)
    rates_variance_over_channel = np.var(rates)
    return {
        "rates": rates,
        "mean": rates_mean_over_channel,
        "variance": rates_variance_over_channel,
    }


@dataclass
class MFRComparison(OperatorMixin):
    recording_duration: float = None
    tag: str = "Mean Firing Rate Comparison"
    channels: List[int] = None

    def __call__(self, pre_spiketrains: Spikestamps, post_spiketrains: Spikestamps):
        assert (
            pre_spiketrains.number_of_channels == post_spiketrains.number_of_channels
        ), f"Number of channels does not match: {pre_spiketrains.number_of_channels} vs {post_spiketrains.number_of_channels}"
        pre_rates = firing_rates(pre_spiketrains)["rates"]
        post_rates = firing_rates(post_spiketrains)["rates"]
        if self.recording_duration is None:
            self.recording_duration = max(
                pre_spiketrains.get_last_spikestamp(),
                post_spiketrains.get_last_spikestamp(),
            ) - min(
                pre_spiketrains.get_first_spikestamp(),
                post_spiketrains.get_first_spikestamp(),
            )
        return pre_rates, post_rates

    def __post_init__(self):
        super().__init__()

    def plot_mfr_comparison(self, output, show=False, save_path=None):
        MFR_pre, MFR_post = output

        if self.channels is not None:
            MFR_pre = MFR_pre[self.channels]
            MFR_post = MFR_post[self.channels]

        MFR = np.geomspace(1e-1, 1e2)
        kl = 7
        ku = 7
        sigma = (
            np.sqrt(np.mean([MFR_pre, MFR_post]) * self.recording_duration)
            / self.recording_duration
        )

        fig = plt.figure()
        plt.plot(MFR, MFR, "b--")
        plt.plot(MFR, MFR - kl * sigma, "b")
        plt.plot(MFR, MFR + ku * sigma, "b")
        plt.scatter(MFR_pre, MFR_post)
        ax = plt.gca()
        ax.set_yscale("log")
        ax.set_xscale("log")
        ax.set_xlim([1e-1, 1e2])
        ax.set_ylim([1e-1, 1e2])
        ax.set_xlabel("MFR pre")
        ax.set_ylabel("MFR post")

        if show:
            plt.show()
        if save_path is not None:
            fig.savefig(os.path.join(f"{save_path}", "mfr.png"))

        return ax


def interspike_intervals(spikes: SpikestampsType):
    """
    Compute the inter-spike intervals of the given spike train.

    Examples
    --------

    How to draw Inter-spike interval histogram (ISIH):

        >>> from miv.statistics import interspike_intervals
        >>> import matplotlib.pyplot as plt
        >>> interspike_interval = interspike_intervals(spikestamp)
        >>> plt.hist(interspike_interval)

    If one wants to get the bin-count, you can use `numpy.digitize` and `numpy.bincount`:

        >>> import numpy as np
        >>> max_time = spikestamps.max()
        >>> num_bins = 20
        >>> time_interval = np.linspace(0, max_time, num_bins)
        >>> digitized = np.digitize(interspike_interval, time_interval)
        >>> bin_counts = np.bincount(digitized)

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        interval: numpy.ndarray

    """

    spike_interval = np.diff(spikes)
    return spike_interval


def coefficient_variation(self, spikes: SpikestampsType):
    """
    The Coefficient of variation: ratio of interspike standard deviation over mean.

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        coefficient variation : float
    """
    interspike = self.interspike_intervals()
    return np.std(interspike) / np.mean(interspike)


def fano_factor(
    spiketrains: SpikestampsType,
    bin_size: float = 0.002,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Calculates the Fano factor for given signal by dividing it into the specified number of bins

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    bin_size : int
        Size of the bin in second. Default is 2ms
    t_start : float
        Binning start time
    t_end : float
        Binning end time

    Returns
    -------
        fano_fac: float
        fanofactor for the specified channel and conditions

    """
    bin_spike = spiketrains.binning(
        bin_size=bin_size, t_start=t_start, t_end=t_end, return_count=True
    )
    fano_factor = np.zeros(bin_spike.number_of_channels, dtype=np.float_)
    for channel in range(bin_spike.number_of_channels):
        array = bin_spike[channel]
        if np.sum(array) == 0:
            fano_factor[channel] = np.nan
        fano_factor[channel] = np.var(array) / np.mean(array)
    return fano_factor


def decay_spike_counts(
    spiketrain, probe_times, amplitude=1.0, decay_rate=5, batchsize=256
):
    """
    decay_spike_counts.

    Both spiketrain and probe_times should be a 1-d array representing time.

    Parameters
    ----------
    spiketrain :
        Single-channel spiketrain
    probe_times :
        probe_times
    amplitude :
        amplitude
    decay_rate :
        decay_rate
    batchsize :
        batchsize
    """
    if len(spiketrain) == 0:
        return np.zeros_like(probe_times)
    batchsize = min(spiketrain.shape[0], batchsize)
    num_sections = spiketrain.shape[0] // batchsize + (
        1 if spiketrain.shape[0] % batchsize > 0 else 0
    )

    result = np.zeros_like(probe_times)
    for subspiketrain in np.array_split(spiketrain, num_sections):
        exponent = (
            np.tile(probe_times, (len(subspiketrain), 1)) - subspiketrain[:, None]
        )
        mask = exponent < 0
        exponent[mask] = 0
        decay_count = np.exp(-decay_rate * exponent)
        decay_count[mask] = 0
        result += decay_count.sum(axis=0)
    return result


def spike_counts_with_kernel(spiketrain, probe_times, kernel: Callable, batchsize=32):
    if len(spiketrain) == 0:
        return np.zeros_like(probe_times)
    spiketrain = np.asarray(spiketrain)

    batchsize = min(spiketrain.shape[0], batchsize)
    num_sections = spiketrain.shape[0] // batchsize + (
        1 if spiketrain.shape[0] % batchsize > 0 else 0
    )

    result = np.zeros_like(probe_times)
    for subspiketrain in np.array_split(spiketrain, num_sections):
        exponent = (
            np.tile(probe_times, (len(subspiketrain), 1)) - subspiketrain[:, None]
        )
        mask = exponent < 0
        exponent[mask] = 0
        decay_count = kernel(exponent)
        # decay_count = np.exp(-decay_rate * exponent)
        decay_count[mask] = 0
        result += decay_count.sum(axis=0)
    return result


def binned_spiketrain(self):
    raise DeprecationWarning("Use `timestamps.binning` instead.")



__all__ = ["MEA128"]

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from miv.mea import mea_map, rhd_64, rhs_32


class MEA128:
    def __init__(self, map_key="128_dual_connector_two_64_rhd"):
        """
        OE Arangement:
        (2)rhd_64 ->, (1)rhd_64 <-
        Intan Arangement
        (D)rhs_32 ->, (C)rhs_32 ->, (B)rhs_32 <-, (A)rhs_32 <-

        -> means chip-side on bottom
        """
        rhd_64_1 = rhd_64[::-1, ::-1].copy()
        rhd_64_1[rhd_64_1 != -1] += 64
        rhd_64_2 = rhd_64.copy()
        self.oe_map = np.concatenate([rhd_64_1, rhd_64_2], axis=0)

        rhs_32_1 = rhs_32[::-1, ::-1].copy()
        rhs_32_1[rhs_32_1 != -1] += 32 * 3
        rhs_32_2 = rhs_32[::-1, ::-1].copy()
        rhs_32_2[rhs_32_2 != -1] += 32 * 2
        rhs_32_3 = rhs_32.copy()
        rhs_32_3[rhs_32_3 != -1] += 32 * 1
        rhs_32_4 = rhs_32.copy()
        rhs_32_4[rhs_32_4 != -1] += 32 * 0
        self.intan_map = np.concatenate(
            [rhs_32_1, rhs_32_2, rhs_32_3, rhs_32_4], axis=0
        )

        self.mea = mea_map[map_key]
        self.mea_intan = np.zeros_like(self.mea, dtype=np.int_) - 1
        for channel in range(128):
            oe_channel = self.channel_mapping(channel, reverse=True)
            self.mea_intan[self.mea == oe_channel] = channel

    def channel_mapping(self, channel, reverse=False):
        """

        Parameters
        ----------
        channel : int or str
        reverse : bool
            If false, mapping is from "given OE channel, output Intan channel"
            If true, mapping is from "given Intan channel, output OE channel"
        """
        if reverse:
            oe = self.oe_map[self.intan_map == channel][0]
            return oe
        else:
            intan = self.intan_map[self.oe_map == channel][0]
            return intan

    def intan_channel_int_to_str(self, intan_channel):
        return chr((intan_channel // 32) + ord("A")) + "-" + str(intan_channel % 32)

    def intan_channel_str_to_int(self, intan_channel):
        group, channel = intan_channel.split("-")
        return (ord(group) - ord("A")) * 32 + int(channel)

    def plot_network(self, channels, show=False, save_path=None, reverse=False):
        """
        OE Arangement:
        (2)rhd_64 ->, (1)rhd_64 <-
        Intan Arangement
        (D)rhs_32 ->, (C)rhs_32 ->, (B)rhs_32 <-, (A)rhs_32 <-

        -> means chip-side on bottom

        Parameters
        ----------
        channel : int or str
        reverse : bool
            If false, mapping is from "given OE channel, output Intan channel"
            If true, mapping is from "given Intan channel, output OE channel"
        """

        # Network Mapping
        oe_arrangement = np.rot90(self.oe_map)
        intan_arrangement = np.rot90(self.intan_map)
        mea = self.mea

        # DEBUG
        # print(oe_arrangement)
        # print(intan_arrangement)
        # print(mea)

        if reverse:
            channels = [self.intan_channel_str_to_int(channel) for channel in channels]

        G = nx.Graph()

        for idl_channel in range(128):
            # if idl_channel in channels:
            #    continue
            idl_intan_channel = self.channel_mapping(idl_channel)

            oe_x, oe_y = np.where(oe_arrangement == idl_channel)
            oe_x, oe_y = oe_x[0], oe_y[0]
            intan_x, intan_y = np.where(intan_arrangement == idl_intan_channel)
            intan_x, intan_y = intan_x[0], intan_y[0]
            mea_x, mea_y = np.where(mea == idl_channel)
            mea_x, mea_y = mea_x[0], mea_y[0]

            G.add_node(
                ("Idel_oe", idl_channel),
                name=str(idl_channel),
                pos=(oe_x, oe_y),
                color="lightgrey",
                alpha=0.5,
            )
            G.add_node(
                ("Idel_intan", idl_intan_channel),
                name=self.intan_channel_int_to_str(idl_intan_channel),
                pos=(intan_x + oe_arrangement.shape[0], intan_y + 0.3),
                color="lightgrey",
                alpha=0.5,
            )
            G.add_node(
                ("Idel_mea", idl_channel),
                name=str(idl_channel),
                pos=(
                    mea_x + oe_arrangement.shape[0] + intan_arrangement.shape[0],
                    mea_y + 0.6,
                ),
                color="lightgrey",
                alpha=0.5,
            )

        for channel in channels:
            if reverse:
                intan_channel = channel
                channel = self.channel_mapping(intan_channel, reverse=True)
            else:
                intan_channel = self.channel_mapping(channel)

            oe_x, oe_y = np.where(oe_arrangement == channel)
            oe_x, oe_y = oe_x[0], oe_y[0]
            intan_x, intan_y = np.where(intan_arrangement == intan_channel)
            intan_x, intan_y = intan_x[0], intan_y[0]
            mea_x, mea_y = np.where(mea == channel)
            mea_x, mea_y = mea_x[0], mea_y[0]
            G.add_node(
                ("OE", channel),
                name=str(channel),
                pos=(oe_x, oe_y),
                color="blue",
                alpha=1.0,
            )
            G.add_node(
                ("Intan", intan_channel),
                name=self.intan_channel_int_to_str(intan_channel),
                pos=(intan_x + oe_arrangement.shape[0], intan_y + 0.3),
                color="red",
                alpha=1.0,
            )
            G.add_node(
                ("MEA", channel),
                name=str(channel),
                pos=(
                    mea_x + oe_arrangement.shape[0] + intan_arrangement.shape[0],
                    mea_y + 0.6,
                ),
                color="grey",
                alpha=1.0,
            )
            G.add_edge(("OE", channel), ("Intan", intan_channel))
            G.add_edge(("Intan", intan_channel), ("MEA", channel))
        pos = {node: (data[1], -data[0]) for node, data in G.nodes(data="pos")}
        labels = {node: data for node, data in G.nodes(data="name")}
        colors = [data for node, data in G.nodes(data="color")]
        alphas = [data for node, data in G.nodes(data="alpha")]

        plt.figure(figsize=(8, 16))
        nx.draw(G, pos=pos, node_size=500, node_color=colors, alpha=alphas)
        nx.draw_networkx_labels(G, pos, labels, font_size=8, font_color="white")
        nx.draw_networkx_edges(G, pos, width=2, edge_color="black", alpha=0.8)

        if save_path is not None:
            plt.savefig(save_path)
        if show:
            plt.show()

        return


if __name__ == "__main__":
    mea_128 = MEA128()

    # targets = [0,11,22,33,44,55,66,77,88,99,101,111]
    # mea_128.plot_network(targets)

    targets = [
        "B-21",
        "A-25",
        "A-29",
        "A-27",
        "A-31",
        "A-26",
        "A-4",
        "C-24",
        "C-9",
        "C-6",
        "C-4",
        "C-16",
        "C-14",
        "D-28",
        "D-31",
        "D-20",
        "D-5",
        "D-14",
        "D-4",
    ]
    mea_128.plot_network(targets, reverse=True, show=True)



from typing import Any, Union

from miv.core.datatype.events import *
from miv.core.datatype.protocol import *
from miv.core.datatype.pure_python import *
from miv.core.datatype.signal import *
from miv.core.datatype.spikestamps import *

DataTypes = Any  # Union[ # TODO
#    miv.core.datatype.signal.Signal,
#    miv.core.datatype.signal.Spikestamps,
# ]



__doc__ = """
Connectivity module
"""
__all__ = ["DirectedConnectivity"]  # , 'UndirectedConnectivity']

from typing import Any, List, Optional, Union

import functools
import gc
import glob
import itertools
import logging
import multiprocessing as mp
import os
import pathlib
import pickle as pkl
from dataclasses import dataclass

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pyinform
import pyinform.transferentropy as pyte
import quantities as pq
import scipy.stats as spst
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.policy import InternallyMultiprocessing
from miv.core.wrapper import wrap_cacher
from miv.mea import mea_map


# self MPI-able
@dataclass
class DirectedConnectivity(OperatorMixin):
    """
    Directional connectivity analysis operator

    Parameters
    ----------
    bin_size : pq.Quantity
        Bin size for spike train
    mea: Optional[Union[str, np.ndarray]]
        2D array map of MEA channel layout.
    tag : str, optional
        Tag for this operator, by default "directional connectivity analysis"
    progress_bar : bool, optional
        Show progress bar, by default True
    skip_surrogate : bool, optional
        Skip surrogate analysis, by default False
    surrogate_N : int, optional
        Number of surrogate, by default 30
    p_threshold : float, optional
        p-value threshold, by default 0.05
    seed : int, optional
        Random seed. If None, use random seed, by default None
    """

    mea: str = None
    channels: Optional[List[int]] = None
    bin_size: float = 0.001
    tag: str = "directional connectivity analysis"
    progress_bar: bool = False

    skip_surrogate: bool = False

    # Surrogate parameters
    surrogate_N: int = 30
    p_threshold: float = 0.05
    H_threshold: float = 1e-2
    seed: int = None
    num_proc: int = 1

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    @wrap_cacher(cache_tag="adjacency_matrix")
    def __call__(self, spikestamps: Spikestamps) -> np.ndarray:
        """__call__.

        Parameters
        ----------
        spikestamps : Spikestamps
        """

        binned_spiketrain: Signal = spikestamps.binning(bin_size=self.bin_size)

        # Channel Selection
        if self.channels is None:
            n_nodes = binned_spiketrain.number_of_channels
            channels = tuple(range(n_nodes))
        else:
            n_nodes = len(self.channels)
            channels = tuple(self.channels)
            binned_spiketrain = binned_spiketrain.select(channels)

        # Get adjacency matrix based on transfer entropy
        adj_matrix = np.zeros([n_nodes, n_nodes], dtype=np.bool_)  # source -> target
        connectivity_metric_matrix = np.zeros(
            [n_nodes, n_nodes], dtype=np.float_
        )  # source -> target

        pairs = [(i, j) for i, j in itertools.product(range(n_nodes), range(n_nodes))]
        func = functools.partial(
            self._get_connection_info,
            binned_spiketrain=binned_spiketrain,
            channels=channels,
            mea=self.mea_map,
            skip_surrogate=self.skip_surrogate,
            surrogate_N=self.surrogate_N,
            seed=self.seed,
            H_threshold=self.H_threshold,
        )
        with mp.Pool(self.num_proc) as pool:
            for idx, result in enumerate(
                tqdm(
                    pool.imap(func, pairs),
                    total=len(pairs),
                    disable=not self.progress_bar,
                )
            ):
                i, j = pairs[idx]
                adj_matrix[i, j] = result[0] < self.p_threshold
                connectivity_metric_matrix[i, j] = result[1]

        connection_ratio = adj_matrix.sum() / adj_matrix.ravel().shape[0]

        info = dict(
            adjacency_matrix=adj_matrix,
            connectivity_matrix=connectivity_metric_matrix,
            connection_ratio=connection_ratio,
        )

        return info

    @staticmethod
    def _surrogate_t_test(
        source, target, skip_surrogate=False, surrogate_N=30, seed=None
    ):
        """
        Surrogate t-test
        """
        # Function configuration. TODO: Make this dependency injection
        te_history = 21
        sublength = 64
        stride = 8

        assert (
            source.shape[0] == target.shape[0]
        ), f"source.shape={source.shape}, target.shape={target.shape}"
        assert (
            source.shape[0] - sublength > 0
        ), f"source.shape[0]={source.shape[0]}, sublength={sublength}"

        rng = np.random.default_rng(seed)  # TODO take rng instead

        func = pyte.transfer_entropy
        te = func(source, target, te_history)
        normalizer = pyinform.entropyrate.entropy_rate(
            target, k=te_history, local=False
        )
        if np.isclose(normalizer, 0.0):
            return 0.0, [0.0]
        te = te / normalizer

        surrogate_tes = []
        if skip_surrogate:
            return te, surrogate_tes
        for _ in range(surrogate_N):
            surrogate_source = source.copy()
            rng.shuffle(surrogate_source)
            for start_index in np.arange(0, source.shape[0] - sublength, stride):
                end_index = start_index + sublength
                surr_te = func(
                    surrogate_source[start_index:end_index],
                    target[start_index:end_index],
                    te_history,
                )
                surrogate_tes.append(surr_te / normalizer)

        return te, surrogate_tes

    @staticmethod
    def _get_connection_info(
        pair,
        binned_spiketrain,
        channels,
        mea,
        skip_surrogate,
        surrogate_N,
        seed,
        H_threshold,
    ):
        """
        Get connection information
        """
        sid, tid = pair
        if sid == tid:
            return 1, 0
        if channels[sid] not in mea or channels[tid] not in mea:
            return 1, 0
        source = binned_spiketrain[sid]
        target = binned_spiketrain[tid]
        te, surrogate_te_list = DirectedConnectivity._surrogate_t_test(
            source,
            target,
            skip_surrogate=skip_surrogate,
            surrogate_N=surrogate_N,
            seed=seed,
        )
        if te < H_threshold:
            return 1, 0
        if skip_surrogate:
            return 1, te
        t_value, p_value = spst.ttest_1samp(surrogate_te_list, te, nan_policy="omit")
        return p_value, te

    def plot_adjacency_matrix(self, result, save_path=None, show=False):
        connectivity_metric_matrix = result["connectivity_matrix"]

        fig, ax = plt.subplots(figsize=(12, 12))
        im = ax.imshow(connectivity_metric_matrix, cmap="gray_r", vmin=0, vmax=1)
        ax.set_xlabel("Source")
        ax.set_ylabel("Target")
        ax.set_title("Transfer Entropy")
        plt.colorbar(im, ax=ax)

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te.png"))

        if show:
            plt.show()

        plt.close(fig)

    def plot_transfer_entropy_histogram(self, result, save_path=None, show=False):
        connectivity_metric_matrix = result["connectivity_matrix"]

        fig, ax = plt.subplots(figsize=(12, 12))
        ax.hist(connectivity_metric_matrix, bins=100)
        ax.set_xlabel("transfer entropy")
        ax.set_ylabel("count")
        ax.set_title("Transfer Entropy Histogram")
        ax.set_xlim([-0.1, 1.1])

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te_histogram.png"))

        if show:
            plt.show()

        plt.close(fig)

    def plot_nodewise_connectivity(
        self,
        result: Any,
        save_path: Union[str, pathlib.Path] = None,
        show: bool = False,
    ):
        """
        Plot nodewise connectivity

        Parameters
        ----------
        result : Any
            Result of this operation.
        save_path : str, optional
            Save path, by default None
        show : bool, optional
            Show plot, by default False

        """
        if show:
            logging.warning(
                "show is not supported for node-wise connectivity plot. Plots will be saved only, if save_path is specified."
            )
            show = False
        if save_path is None:
            return
        adj_matrix = result["adjacency_matrix"]
        connectivity_metric_matrix = result["connectivity_matrix"]

        n_nodes = adj_matrix.shape[0]
        channels = self.channels if self.channels is not None else list(range(n_nodes))

        # Save connectivity plot
        for idx, source in enumerate(channels):
            if source not in self.mea_map:
                continue
            self._plot_directionality(
                adj_matrix[idx],
                source,
                os.path.join(save_path, f"p_graph_source_{source}.png"),
            )
            self._plot_directionality(
                connectivity_metric_matrix[idx],
                source,
                os.path.join(save_path, f"te_graph_source_{source}.png"),
                boolean_connectivity=False,
            )

        for idx, target in enumerate(channels):
            if target not in self.mea_map:
                continue
            self._plot_directionality(
                adj_matrix[:, idx],
                target,
                os.path.join(save_path, f"p_graph_target_{target}.png"),
                reverse=True,
            )
            self._plot_directionality(
                connectivity_metric_matrix[:, idx],
                target,
                os.path.join(save_path, f"te_graph_target_{target}.png"),
                reverse=True,
                boolean_connectivity=False,
            )

    def _plot_directionality(
        self,
        connectivity,
        self_index,
        savename,
        reverse=False,
        boolean_connectivity=True,
    ):
        """
        Plot directionality
        """
        G_1 = nx.DiGraph()

        # position
        pos = {}
        for i in range(self.mea_map.shape[0]):
            for j in range(self.mea_map.shape[1]):
                center = self.mea_map[i, j]
                if center < 0:
                    continue
                G_1.add_node(center)
                pos[center] = (j, -i)

        if boolean_connectivity:
            for idx, conn in enumerate(connectivity):
                if idx == self_index:
                    continue  # No self-connection
                if idx not in self.mea_map:
                    continue  # No connection if node does not exist
                if not conn:
                    continue  # No connection

                if not reverse:
                    G_1.add_edge(self_index, idx)
                else:
                    G_1.add_edge(idx, self_index)
        else:
            for idx, conn in enumerate(connectivity):
                if idx == self_index:
                    continue  # No self-connection
                if idx not in self.mea_map:
                    continue  # No connection if node does not exist

                if not reverse:
                    G_1.add_edge(self_index, idx, weight=conn)
                else:
                    G_1.add_edge(idx, self_index, weight=conn)

        # Numbering
        # edge_weights = [G_1[u][v]['weight'] if 'weight' in G_1[u][v] else 0.0 for u, v in G_1.edges()]
        plt.figure()
        widths = nx.get_edge_attributes(G_1, "weight")
        nx.draw(G_1, pos)
        if boolean_connectivity:
            nx.draw_networkx_edges(
                G_1,
                pos,
                edgelist=widths.keys(),
                ax=plt.gca(),
            )
        else:
            nx.draw_networkx_edges(
                G_1,
                pos,
                edgelist=widths.keys(),
                edge_color=list(widths.values()),
                edge_cmap=plt.cm.gray_r,
                edge_vmin=0.0,
                edge_vmax=1.0,
                ax=plt.gca(),
            )
        nx.draw_networkx_labels(
            G_1,
            pos,
            labels=dict(zip(G_1.nodes(), G_1.nodes())),
            font_color="white",
            ax=plt.gca(),
        )
        # for node, (x_coord, y_coord) in pos.items():
        #    plt.text(x_coord - 0.1, y_coord - 0.1, str(node))

        plt.savefig(savename)
        plt.close()


# TODO
# @dataclass
# class UndirectedConnectivity(OperatorMixin):
#    pass



from typing import Generator, Protocol

from miv.core.datatype.protocol import Extendable


class _Collapsable(Protocol):
    @classmethod
    def from_collapse(self) -> None:
        ...


class CollapseExtendableMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_collapse(cls, values: Generator[Extendable, None, None]):
        obj = cls()
        for value in values:
            obj.extend(value)
        return obj



__doc__ = """ Base mixin for MEA classes. """

from typing import Tuple

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from miv.core.operator import BaseChainingMixin


class MEAMixin(BaseChainingMixin):
    """Base mixin for MEA classes.

    Functional module.

    Comply with the following interface:
        MEAGeometryProtocol
    """

    def __init__(self, tag: str = "mea", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tag = tag
        self.runner = None

    def get_xy(self, idx: int) -> Tuple[float, float]:
        """Given node index, return xy coordinate"""
        raise NotImplementedError

    def save(self, path: str) -> None:
        raise NotImplementedError

    def load(self, path: str) -> None:
        raise NotImplementedError

    def view(self) -> plt.Figure:
        """Simplified view of MEA orientation"""
        raise NotImplementedError

    @property
    def output(self):
        return self

    def run(self, **kwargs):
        pass

    def map_data(self, vector: np.ndarray) -> np.ndarray:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        raise NotImplementedError



__doc__ = """Centrality module"""
__all__ = ["plot_eigenvector_centrality"]

import os

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


def plot_eigenvector_centrality(self, result, show=False, save_path=None):
    metric_matrix = result["connectivity_matrix"]
    adjacency_matrix = result["adjacency_matrix"]
    n_nodes = metric_matrix.shape[0]
    channels = self.channels if self.channels is not None else list(range(n_nodes))

    # def plot_centrality(self, adjacency_matrix, n_nodes, centrality, mea, ax=None, savename=None, include_colorbar:bool=True):
    # def get_graph(adjacency_matrix, n_nodes):

    # Construct graph
    G = nx.DiGraph()

    # position
    pos = {}
    for i in range(self.mea_map.shape[0]):
        for j in range(self.mea_map.shape[1]):
            center = self.mea_map[i, j]
            if center < 0:
                continue
            if center not in channels:
                continue
            G.add_node(center)
            pos[center] = (j, -i)

    for source in range(n_nodes):
        if channels[source] not in self.mea_map:
            continue
        for target in range(n_nodes):
            if channels[target] not in self.mea_map or target not in channels:
                continue
            if source == target:
                conn = 0
            else:
                conn = adjacency_matrix[source, target]
            if np.isnan(conn):
                conn = 0.0
            G.add_edge(channels[source], channels[target], weight=conn)

    # Plotting
    nodes = G.nodes()
    centrality = nx.eigenvector_centrality_numpy(G, weight="weight")
    colors = [centrality[n] for n in nodes]

    vmin = np.min(list(centrality.values()))
    vmax = np.max(list(centrality.values()))
    v = np.max(np.abs([vmin, vmax]))

    fig = plt.figure()
    ax = plt.gca()

    # nx.draw_networkx_edges(G, pos, alpha=0.2, ax=ax)
    nc = nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=nodes,
        node_color=colors,
        node_size=100,
        cmap=plt.cm.RdBu,
        ax=ax,
        vmin=-v,
        vmax=v,
    )
    for node, (x_coord, y_coord) in pos.items():
        ax.text(x_coord - 0.1, y_coord - 0.1, str(node))
    fig.colorbar(nc, ax=ax)

    if save_path is not None:
        plt.savefig(os.path.join(save_path, "eigenvector_centrality.png"))
    if show:
        plt.show()
    plt.close("all")



import logging
import multiprocessing as mp
import os
import sys
from glob import glob

import click
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.core.pipeline import Pipeline
from miv.signal.filter import ButterBandpass
from miv.signal.spike import ThresholdCutoff


@click.command()
@click.option("--path", "-p", help="Path to the recorded directory.", multiple=True)
@click.option(
    "--tools",
    "-t",
    type=click.Choice(["OpenEphys", "Intan"]),
    default="OpenEphys",
    help="Select the recording format",
)
@click.option(
    "--nproc",
    "-N",
    default=None,
    help="Number of processing units. If not provided, use maximum.",
)
@click.option(
    "--num_fragments",
    "-n",
    default=1,
    help="Number of fragments for data processing. Recommend to split data into 1 minute recording segments. For Intan, this number is pre-determined by the number of saved files.",
)
@click.option(
    "--use-mpi",
    type=bool,
    default=False,
    help="Set True if mpi is ready. Else, it will use multiprocessing. (mpi4py must be installed)",
)
@click.option("--chunksize", default=1, help="Number of chunks for multiprocessing")
def main(path, tools, nproc, num_fragments, use_mpi, chunksize):
    signal_filter = ButterBandpass(400, 1500, order=4)
    spike_detection = ThresholdCutoff(cutoff=5.0, progress_bar=True)
    signal_filter >> spike_detection

    # Select tools
    if tools == "OpenEphys":
        # If OpenEphys is selected, process all recordings in the set.
        from miv.io.openephys import DataManager

        for p in path:
            for data in DataManager(p):
                data >> signal_filter
                pipeline = Pipeline(spike_detection)
                pipeline.run(save_path=data.analysis_path, no_cache=True)
                spike_detection.plot(save_path=data.analysis_path)
                logging.info(f"Pre-processing {p}-{data.data_path} done.")
                data.clear_connections()
    elif tools == "Intan":
        from miv.io.intan import DataIntan

        for p in path:
            data = DataIntan(p)
            pipeline = Pipeline(spike_detection)
            pipeline.run(save_path=data.analysis_path, no_cache=True)
            spike_detection.plot(save_path=data.analysis_path)
            logging.info(f"Pre-processing {p}-{data.data_path} done.")
            data.clear_connections()


if __name__ == "__main__":
    main()



import xml.etree.ElementTree as ET

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt

# Parse the XML to get the color map data
xml_data = """
<ColorMaps>
<ColorMap name="New Color Preset" space="Diverging">
  <Point x="-40" o="1" r="0.231373" g="0.298039" b="0.752941"/>
  <Point x="-8" o="1" r="0.535058" g="0.723751" b="0.85594"/>
  <Point x="-1" o="1" r="1" g="1" b="1"/>
  <Point x="0" o="1" r="1" g="1" b="1"/>
  <Point x="1" o="1" r="1" g="1" b="1"/>
  <Point x="8" o="1" r="1" g="0.713756" b="0.294972"/>
  <Point x="40" o="1" r="0.733898" g="0.0134737" b="0.150759"/>
</ColorMap>
</ColorMaps>
"""

cdict = {
    "red": [
        [
            0.0,
        ],
        [
            0.5,
        ],
        [
            1.0,
        ],
    ],
    "green": [
        [
            0.0,
        ],
        [
            0.25,
        ],
        [
            0.75,
        ],
        [
            1.0,
        ],
    ],
    "blue": [
        [
            0.0,
        ],
        [
            0.5,
        ],
        [
            1.0,
        ],
    ],
}


def plot_linearmap(cdict):
    newcmp = LinearSegmentedColormap("testCmap", segmentdata=cdict, N=256)
    rgba = newcmp(np.linspace(0, 1, 256))
    fig, ax = plt.subplots(figsize=(4, 3), constrained_layout=True)
    col = ["r", "g", "b"]
    for xx in [0.25, 0.5, 0.75]:
        ax.axvline(xx, color="0.7", linestyle="--")
    for i in range(3):
        ax.plot(np.arange(256) / 256, rgba[:, i], color=col[i])
    ax.set_xlabel("index")
    ax.set_ylabel("RGB")
    plt.show()


plot_linearmap(cdict)



__doc__ = """"""
__all__ = ["_Callback"]

from typing import TypeVar  # TODO: For python 3.11, we can use typing.Self
from typing import Callable, Optional, Protocol, Union

import itertools
import os
import pathlib

import matplotlib.pyplot as plt

SelfCallback = TypeVar("SelfCallback", bound="_Callback")


def MixinOperators(func):
    return func


@MixinOperators
def get_methods_from_feature_classes_by_startswith_str(self, method_name: str):
    methods = [
        [
            v
            for (k, v) in cls.__dict__.items()
            if k.startswith(method_name) and method_name != k and callable(v)
        ]
        for cls in self.__class__.__mro__
    ]
    return list(itertools.chain.from_iterable(methods))


@MixinOperators
def get_methods_from_feature_classes_by_endswith_str(self, method_name: str):
    methods = [
        [
            v
            for (k, v) in cls.__dict__.items()
            if k.endswith(method_name) and method_name != k and callable(v)
        ]
        for cls in self.__class__.__mro__
    ]
    return list(itertools.chain.from_iterable(methods))


class _Callback(Protocol):
    def __lshift__(self, right: SelfCallback) -> SelfCallback:
        ...

    def receive(self):
        ...

    def output(self):
        ...

    def callback_before_run(self):
        ...

    def callback_after_run(self):
        ...


class BaseCallbackMixin:
    def __init__(self):
        super().__init__()
        self._callback_before_run = []
        self._callback_after_run = []
        self._callback_plot = []
        self.skip_plotting: bool = False

    def __lshift__(self, right: Callable) -> SelfCallback:
        if right.__name__.startswith(
            "__prepend"
        ):  # TODO: need better way to prepend callbacks
            self._callback_before_run.append(right)
            return self
        if right.__name__.startswith("plot_"):
            self._callback_plot.append(right)
            return self
        self._callback_after_run.append(right)
        return self

    def callback_before_run(self, inputs):
        predefined_callbacks = get_methods_from_feature_classes_by_startswith_str(
            self, "before_run"
        )
        for callback in predefined_callbacks + self._callback_before_run:
            inputs = callback(self, inputs)
        return inputs

    def callback_after_run(self, output):
        predefined_callbacks = get_methods_from_feature_classes_by_startswith_str(
            self, "after_run"
        )
        for callback in predefined_callbacks + self._callback_after_run:
            output = callback(self, output)
        return output

    def plot_from_callbacks(self, *args, **kwargs):
        for callback in self._callback_plot:
            callback(self, *args, **kwargs)

    def plot(
        self,
        show: bool = False,
        save_path: Optional[Union[bool, str, pathlib.Path]] = None,
        dry_run: bool = False,
    ):
        if self.skip_plotting:
            return
        if save_path is True:
            os.makedirs(self.analysis_path, exist_ok=True)
            save_path = self.analysis_path
        plotters = get_methods_from_feature_classes_by_startswith_str(self, "plot")
        if dry_run:
            for plotter in plotters:
                print(f"dry run: {plotter}")
            return
        for plotter in plotters:
            plotter(self, self._output, show=show, save_path=save_path)
        if not show:
            plt.close("all")



__all__ = [
    "probability_distribution",
    "shannon_entropy",
    "block_entropy",
    "entropy_rate",
    "active_information",
    "mutual_information",
    "relative_entropy",
    "joint_entropy",
    "conditional_entropy",
    "cross_entropy",
    "transfer_entropy",
    "partial_information_decomposition",
]


from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Union

import datetime

import elephant.statistics
import matplotlib.pyplot as plt
import neo
import numpy as np
import pyinform
import quantities as pq
import scipy
import scipy.signal
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.statistics.spiketrain_statistics import binned_spiketrain

INFO_METRICS_TAG = Literal["self", "pair", "all"]  # Pragma: no cover


def tag_info_metrics(tag: INFO_METRICS_TAG) -> Callable:  # Pragma: no cover
    """
    Decorator to tag the info metrics functions
        - self: channel-wise metrics
        - pair: channel pair-wise metrics
        - all: combined across all channels
    """

    def decorator(func: Callable) -> Callable:
        func.tag = tag
        return func

    return decorator


@tag_info_metrics("self")
def probability_distribution(
    spiketrains: Spikestamps,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):

    """
    Forms the probability distribution required to compute the information theory measures. Probability is computed based on the binned spiketrain generated for the specified bin size.

    Parameters
    ----------
    spiketrains : Spikestamps
    bin_size : float
        bin size in seconds
    t_start : Optional[float]
        Binning start time. If None, the start time is the first spike time.
    t_end : Optional[float]
        Binning end time. If None, the end time is the last spike time.

    Returns
    -------
        probability_distribution: np.ndarray
        probability distribution for the provided spiketrain

    """
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    prob_spike = (
        bin_spike.data.sum(axis=bin_spike._SIGNALAXIS, keepdims=True)
        / bin_spike.shape[bin_spike._SIGNALAXIS]
    )
    probability_distribution = np.empty_like(bin_spike.data, dtype=np.float_)
    _prob_spike = np.repeat(
        prob_spike, bin_spike.shape[bin_spike._SIGNALAXIS], axis=bin_spike._SIGNALAXIS
    )
    probability_distribution[bin_spike.data] = _prob_spike[bin_spike.data]
    probability_distribution[~bin_spike.data] = (1 - _prob_spike)[~bin_spike.data]
    return probability_distribution


@tag_info_metrics("self")
def shannon_entropy(
    spiketrains: Spikestamps,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the shannon entropy for a single channel recording using the binned spiketrain.

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
        shannon_entropy: float
        Shannon entropy for the given channel

    """
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    prob_spike = (
        bin_spike.data.sum(axis=bin_spike._SIGNALAXIS, keepdims=True)
        / bin_spike.shape[bin_spike._SIGNALAXIS]
    ).astype(np.float_)
    prob_no_spike = 1 - prob_spike
    shannon_entropy = -(
        prob_spike * np.log2(prob_spike) + prob_no_spike * np.log2(prob_no_spike)
    )
    return shannon_entropy


@tag_info_metrics("self")
def block_entropy(
    spiketrains: Spikestamps,
    history: int,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the block entropy for a recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : int
        history length
    bin_size : float
        bin size in seconds
    t_start : float
        Binning start time
    t_end : float
        Binning end time

    Returns
    -------
    block_entropy: float
        Block entropy for the given channel

    """
    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    block_entropy = np.zeros((bin_spike.shape[bin_spike._CHANNELAXIS],))
    for idx, bin_spike_channel in enumerate(bin_spike):
        block_entropy[idx] = pyinform.blockentropy.block_entropy(
            bin_spike_channel, history
        )
    return block_entropy


@tag_info_metrics("self")
def entropy_rate(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the entropy rate for a each channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : int
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    entropy rate: float
        entropy rate for the given channel

    """
    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    entropy_rate = []
    for idx, bin_spike_channel in enumerate(bin_spike):
        entropy_rate.append(
            pyinform.entropyrate.entropy_rate(bin_spike_channel, k=history, local=True)[
                0
            ]
        )
    return np.asarray(entropy_rate), bin_spike  # TODO: return single type


@tag_info_metrics("self")
def active_information(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the active information for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    bin_size : float
        bin size in seconds
    history : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time

    Returns
    -------
    active_information: float
        Active information for the given channel

    """
    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    active_information = np.zeros((bin_spike.shape[bin_spike._CHANNELAXIS],))
    for idx, bin_spike_channel in enumerate(bin_spike):
        active_information[idx] = pyinform.activeinfo.active_info(
            bin_spike_channel, k=history
        )
    return active_information


@tag_info_metrics("pair")
def transfer_entropy(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the transfer entropy for the pair of electorde recordings (X & Y) using the binned spiketrains and history

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    transfer_entropy: float
        Transfer_entropy for the given pair of electrodes

    """
    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)

    transfer_entropy_matrix = np.zeros(
        [bin_spike.number_of_channels, bin_spike.number_of_channels]
    )
    for channelx in tqdm(range(bin_spike.number_of_channels)):
        for channely in tqdm(range(bin_spike.number_of_channels), leave=False):
            bin_spike_x = bin_spike[channelx]
            bin_spike_y = bin_spike[channely]

            # np.random.shuffle(bin_spike_x)

            normalizer = pyinform.entropyrate.entropy_rate(
                bin_spike_y, k=history, local=False
            )
            # normalizer = pyinform.mutualinfo.mutual_info(bin_spike_x, bin_spike_y, local=False)

            transfer_entropy = pyinform.transferentropy.transfer_entropy(
                bin_spike_x, bin_spike_y, k=history, local=False
            )
            if np.isclose(normalizer, 0):
                transfer_entropy_matrix[channelx, channely] = 0
            else:
                transfer_entropy_matrix[channelx, channely] = (
                    transfer_entropy / normalizer
                )
    return transfer_entropy_matrix


@tag_info_metrics("pair")
def mutual_information(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the mutual information for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    mutual_information: float
        Mutual information for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"
    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    mutual_information = pyinform.mutualinfo.mutual_info(bin_spike_x, bin_spike_y)
    return mutual_information


@tag_info_metrics("pair")
def joint_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the joint entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    joint_entropy: float
        joint entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_xy = np.logical_and(spike_dist_x, spike_dist_y)
    joint_entropy = -np.sum(spike_dist_xy * np.log2(spike_dist_xy))
    return joint_entropy


@tag_info_metrics("pair")
def relative_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the relative entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    relative_entropy: float
        Relative_entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    relative_entropy = pyinform.relativeentropy.relative_entropy(
        bin_spike_x, bin_spike_y
    )
    return relative_entropy


@tag_info_metrics("pair")
def conditional_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the conditional entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    conditional_entropy: float
        conditional entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    conditional_entropy = pyinform.conditionalentropy.conditional_entropy(
        bin_spike_x, bin_spike_y
    )
    return conditional_entropy


@tag_info_metrics("pair")
def cross_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Estimates the cross entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    cross_entropy: float
        cross entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    cross_entropy = -np.sum(spike_dist_x * np.log2(spike_dist_y))
    return cross_entropy


@tag_info_metrics("pair")
def partial_information_decomposition(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    channelz: float,
    bin_size: float,
    t_start: Optional[float] = None,
    t_end: Optional[float] = None,
):
    """
    Decomposes the information provided by channel x and y about channel z in redundancy, unique information, and synergy.

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    channelz : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    redundancy: float
        redundant information provided by both x and y about z
    unique_information_x: float
        information uniquely provided by x about z
    unique_information_y: float
        information uniquely provided by y about z
    synergy: float
        synergetic information provided by both x and y about z
    total_information: float
        total information provided by both x and y about z
    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_z = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_xz = np.logical_and(spike_dist_x, spike_dist_z)
    spike_dist_yz = np.logical_and(spike_dist_y, spike_dist_z)
    spike_dist_xy = np.logical_and(spike_dist_x, spike_dist_y)
    spike_dist_xyz = np.logical_and(spike_dist_x, spike_dist_y, spike_dist_z)
    I_x_z = np.sum(
        spike_dist_xz * np.log2(spike_dist_xz / (spike_dist_x * spike_dist_z))
    )
    I_y_z = np.sum(
        spike_dist_yz * np.log2(spike_dist_yz / (spike_dist_y * spike_dist_z))
    )
    redundancy = np.min(I_x_z, I_y_z)
    unique_information_x = I_x_z - redundancy
    unique_information_y = I_y_z - redundancy
    total_information = np.sum(
        spike_dist_xyz * np.log2(spike_dist_xyz / (spike_dist_xy * spike_dist_z))
    )
    synergy = (
        total_information - redundancy - unique_information_x - unique_information_y
    )
    return (
        redundancy,
        unique_information_x,
        unique_information_y,
        synergy,
        total_information,
    )



__doc__ = """

Module (Intan)
##################

.. autoclass:: DataIntan
   :members:

.. autoclass:: DataIntanTriggered
   :members:

"""
__all__ = ["DataIntan", "DataIntanTriggered"]

from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import logging
import os
import pickle
import xml.etree.ElementTree as ET
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

import miv.io.intan.rhs as rhs
from miv.core.datatype import Signal, Spikestamps
from miv.core.wrapper import wrap_cacher
from miv.io.openephys.data import Data, DataManager
from miv.typing import SignalType


class DataIntan(Data):
    """Single data unit handler, collected from Intan.

    Each data unit that contains single recording. This class provides useful tools,
    such as masking channel, export data, interface with other packages, etc.
    If you have multiple recordings you would like to handle at the same time, use
    `DataManager` instead.

    By default recording setup, the following directory structure is expected in ``data_path``::

        data_<date>_<time>              # <- recorded data path
        ├── settings.xml
        ├── data_<date>_<time1>.rhs      # <- first rhs file
        ├── data_<date>_<time2>.rhs      # <- second rhs file
        ├── data_<date>_<time3>.rhs      # <- third rhs file
        ⋮
        └── analysis

        Parameters
        ----------
        data_path : str
    """

    def load(self):
        """
        Iterator to load data fragmentally.
        This function loads each file separately.

        Examples
        --------
            >>> data = Data(data_path)
            >>> for data.load(10) as (signal, timestamps, sampling_rate):
            ...     ...

        Returns
        -------
        signal : SignalType, neo.core.AnalogSignal
            The length of the first axis `signal.shape[0]` correspond to the length of the
            signal, while second axis `signal.shape[1]` correspond to the number of channels.
        timestamps : TimestampsType, numpy array
        sampling_rate : float

        Raises
        ------
        FileNotFoundError
            If some key files are missing.
        """

        yield from self._generator_by_channel_name("amplifier_data")

    def get_stimulation(self, progress_bar=False):
        """
        Load stimulation recorded data.
        """
        signals, timestamps = [], []
        sampling_rate = None
        for data in self._generator_by_channel_name("stim_data", progress_bar):
            signals.append(data.data)
            timestamps.append(data.timestamps)
            sampling_rate = data.rate

        return Signal(
            data=np.concatenate(signals, axis=0),
            timestamps=np.concatenate(timestamps),
            rate=sampling_rate,
        )

    def _generator_by_channel_name(self, name: str, progress_bar: bool = False):
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")
        files = self.get_recording_files()
        # Get sampling rate from setting file
        setting_path = os.path.join(self.data_path, "settings.xml")
        sampling_rate = int(ET.parse(setting_path).getroot().attrib["SampleRateHertz"])
        active_channels, total = self._get_active_channels()
        # Read each files
        for filename in tqdm(files, disable=not progress_bar):
            result, data_present = rhs.load_file(filename)
            assert data_present, f"Data does not present: {filename=}."
            assert not hasattr(result, name), f"No {name} in the file ({filename=})."

            signal = np.asarray(result[name])
            if total != signal.shape[0]:
                _signal = np.zeros([total, signal.shape[1]], dtype=signal.dtype)
                _signal[active_channels, :] = signal
                signal = _signal

            yield Signal(
                data=signal.T,
                timestamps=np.asarray(result["t"]),
                rate=sampling_rate,
            )

    def _get_active_channels(self, group_prefix=("A", "B", "C", "D")):
        setting_path = os.path.join(self.data_path, "settings.xml")
        root = ET.parse(setting_path).getroot()
        total = 0
        active_channels = []
        for child in root:
            if child.tag != "SignalGroup":
                continue
            if child.attrib["Prefix"] not in group_prefix:
                continue
            for channel in child:
                if channel.attrib["Enabled"] == "True":
                    active_channels.append(total)
                total += 1
        return np.array(active_channels), total

    def check_path_validity(self):
        """
        Check if necessary files exist in the directory.

        - Check `rhs` file exists.
        - Check `settings.xml` file exists.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.
        """

        continuous_dat_paths = self.get_recording_files()
        if len(continuous_dat_paths) == 0:
            logging.warning("At least one .rhs file must exist in the data path.")
            return False
        if not os.path.exists(os.path.join(self.data_path, "settings.xml")):
            logging.warning("Missing settings.xml in the data path.")
            return False
        return True

    def get_recording_files(self):
        paths = glob(os.path.join(self.data_path, "*.rhs"), recursive=True)
        paths.sort()
        return paths

    def get_stimulation_events(self):  # TODO: refactor
        minimum_stimulation_length = 0.010
        data = self.get_stimulation()
        stim = data.data
        timestamps = data.timestamps
        # sampling_rate = data.rate
        stimulated_channels = np.where(np.abs(stim).sum(axis=0))[0]
        if len(stimulated_channels) == 0:
            return None
        stimulated_channel = stimulated_channels[0]
        stim = stim[:, stimulated_channel]

        events = ~np.isclose(stim, 0)
        eventstrain = timestamps[np.where(events)[0]]
        ref = np.concatenate(
            [[True], np.diff(eventstrain) > minimum_stimulation_length]
        )
        eventstrain = eventstrain[ref]
        ret = Spikestamps([eventstrain])  # TODO: use datatype.Events
        return ret

    # Disable
    def load_ttl_event(self):
        raise AttributeError("DataIntan does not have laod_ttl_event method")


class DataIntanTriggered(DataIntan):
    """
    DataIntanTriggered is a subclass of DataIntan, which is used to handle Intan
    recording when the recording is triggered by TTL signals.
    """

    def __init__(
        self,
        data_path,  # FIXME: argument order with intan.DATA
        index: int = 0,
        trigger_key: str = "board_adc_data",
        trigger_index: int = 0,
        trigger_threshold_voltage=1.0,
        progress_bar: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(data_path=data_path, *args, **kwargs)
        self.index = index
        self.trigger_key = trigger_key
        self.trigger_index = trigger_index
        self.trigger_threshold_voltage = trigger_threshold_voltage
        self.progress_bar = progress_bar

    def __len__(self):
        groups = self._trigger_grouping()
        return len(groups)

    def __getitem__(self, index):
        groups = self._trigger_grouping()
        if len(groups) <= index:
            raise IndexError(
                f"Index exceeds the number of triggered recordings ({len(groups)})."
            )
        return DataIntanTriggered(
            data_path=self.data_path,
            index=index,
            trigger_key=self.trigger_key,
            trigger_index=self.trigger_index,
            trigger_threshold_voltage=self.trigger_threshold_voltage,
            progress_bar=self.progress_bar,
        )

    @wrap_cacher(cache_tag="trigger_grouping")
    def _trigger_grouping(self):
        def _find_sequence(arr):
            if arr.size == 0:
                return arr
            return arr[np.concatenate([np.array([True]), arr[1:] - 1 != arr[:-1]])]

        paths = DataIntan.get_recording_files(self)

        group_files = []
        group = {"paths": [], "start index": [], "end index": []}
        status = 0
        for file in tqdm(paths, disable=not self.progress_bar):
            result, _ = rhs.load_file(file)
            time = result["t"]
            adc_data = result[self.trigger_key][self.trigger_index]
            diff_adc_data = adc_data[1:] - adc_data[:-1]
            # TODO: Something can be done better
            recording_on = _find_sequence(
                np.where(diff_adc_data > self.trigger_threshold_voltage)[0]
            ).tolist()
            recording_off = _find_sequence(
                np.where(diff_adc_data < -self.trigger_threshold_voltage)[0]
            ).tolist()

            sindex = 0
            if status == 1 and len(recording_on) == 0 and len(recording_off) == 0:
                group["paths"].append(file)
                group["start index"].append(sindex)
                group["end index"].append(time.shape[0])
            while len(recording_on) > 0 or len(recording_off) > 0:
                if status == 0 and len(recording_on) > 0:
                    sindex = recording_on.pop(0)
                    status = 1
                    if len(recording_off) == 0:
                        group["paths"].append(file)
                        group["start index"].append(sindex)
                        group["end index"].append(time.shape[0])
                elif status == 1 and len(recording_off) > 0:
                    eindex = recording_off.pop(0)
                    status = 0
                    if eindex < sindex:
                        raise ValueError(
                            f"Something went wrong with the trigger signal. Starting index {sindex} must be less than ending index {eindex}."
                        )
                    group["paths"].append(file)
                    group["start index"].append(sindex)
                    group["end index"].append(eindex)
                    group_files.append(group)
                    group = {"paths": [], "start index": [], "end index": []}
                else:
                    raise ValueError(
                        f"Something went wrong with the trigger signal. {len(recording_on)=} {len(recording_off)=}"
                    )

        return group_files

    def get_recording_files(self):
        groups = self._trigger_grouping()
        return groups[self.index]["paths"]

    def _generator_by_channel_name(self, name: str, progress_bar: bool = False):
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")
        groups = self._trigger_grouping()
        files = groups[self.index]["paths"]
        sindex = groups[self.index]["start index"]
        eindex = groups[self.index]["end index"]
        # Get sampling rate from setting file
        setting_path = os.path.join(self.data_path, "settings.xml")
        sampling_rate = int(ET.parse(setting_path).getroot().attrib["SampleRateHertz"])
        active_channels, total = self._get_active_channels()
        # Read each files
        for filename, sidx, eidx in tqdm(
            zip(files, sindex, eindex), disable=not progress_bar
        ):
            result, data_present = rhs.load_file(filename)
            assert data_present, f"Data does not present: {filename=}."
            assert not hasattr(result, name), f"No {name} in the file ({filename=})."

            signal = np.asarray(result[name])
            if total != signal.shape[0]:
                _signal = np.zeros([total, signal.shape[1]], dtype=signal.dtype)
                _signal[active_channels, :] = signal
                signal = _signal

            yield Signal(
                data=signal.T[sidx:eidx, :],
                timestamps=np.asarray(result["t"])[sidx:eidx],
                rate=sampling_rate,
            )



__doc__ = """
Spikestamps
===========

.. autoclass:: Spikestamps
   :members:

"""

__all__ = ["Spikestamps"]

from typing import List, Optional

from collections import UserList

import numpy as np
import quantities as pq

from miv.core.datatype.collapsable import CollapseExtendableMixin
from miv.core.datatype.signal import Signal
from miv.core.operator.operator import DataNodeMixin


class Spikestamps(UserList, CollapseExtendableMixin, DataNodeMixin):
    """List of array of spike times

    Represents spikes emitted by the same unit in a period of times.
    """

    def __init__(self, iterable: Optional[List] = None):
        super().__init__()
        if iterable is None:
            iterable = []
        self.data = iterable

    @property
    def number_of_channels(self) -> int:
        return len(self.data)

    def __setitem__(self, index, item):
        self.data[index] = item

    def insert(self, index, item):
        if index > len(self.data) or index < 0:
            raise IndexError("Index out of range")
        self.data.insert(index, item)

    def append(self, item):
        self.data.append(item)

    def extend(self, other):
        """
        Extend spikestamps from another `Spikestamps` or list of arrays.

        If the given parameter is another `Spikestamps`, each arrays are concatenated.
        Remaining channels will be added as different channels.

        If the given parameter is list of another arrays, each arrays will be added as different channels.

        Examples
        --------

        >>> a = Spikestamps([[0,1,2],[0,3]])
        >>> b = Spikestamps([[3,5],[4],[0,1]])
        >>> a.extend(b)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1]]

        >>> c = [[1],[5],[0]]
        >>> a.extend(c)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1], [1], [5], [0]]

        """
        if isinstance(other, type(self)):
            length_diff = len(other) - len(self.data)
            if length_diff > 0:
                for _ in range(length_diff):
                    self.data.append([])
            for idx, array in enumerate(other):
                self.data[idx].extend(array)
        else:
            self.data.extend(item for item in other)

    def get_count(self):
        """Return list of spike-counts for each channel."""
        return [len(data) for data in self.data]

    def get_last_spikestamp(self):
        """Return timestamps of the last spike in this spikestamps"""
        return max([max(data) for data in self.data if len(data) > 0])

    def get_first_spikestamp(self):
        """Return timestamps of the first spike in this spikestamps"""
        return min([min(data) for data in self.data if len(data) > 0])

    def get_view(self, t_start: float, t_end: float):
        """Truncate array and only includes spikestamps between t_start and t_end."""
        return Spikestamps(
            [
                np.array(sorted(list(filter(lambda x: t_start <= x <= t_end, arr))))
                for arr in self.data
            ]
        )

    def neo(self):
        """Cast to neo.SpikeTrain"""
        import neo

        t_start = self.get_first_spikestamp()
        t_stop = self.get_last_spikestamp()
        return [
            neo.SpikeTrain(arr, t_start=t_start, t_stop=t_stop, units=pq.s)
            for arr in self.data
        ]

    def binning(
        self,
        bin_size: float = 1 * pq.ms,
        return_count: bool = False,
    ) -> Signal:
        """
        Forms a binned spiketrain using the spiketrain

        Parameters
        ----------
        bin_size : float | pq.Quantity
            bin size in the unit of time.
        return_count : bool
            If set to true, return the bin count. (default=False)

        Returns
        -------
        bin_spike: numpy.ndarray
            binned spiketrain with 1 corresponding to spike and zero otherwise

        """
        spiketrain = self.data

        if isinstance(bin_size, pq.Quantity):
            bin_size = bin_size.rescale(pq.s).magnitude
        assert bin_size > 0, "bin size should be greater than 0"

        t_start = self.get_first_spikestamp()
        t_end = self.get_last_spikestamp()
        n_bins = int(np.ceil((t_end - t_start) / bin_size))
        time = t_start + (np.arange(n_bins + 1) * bin_size)

        num_channels = self.number_of_channels
        signal = Signal(
            data=np.zeros(
                [time.shape[0] - 1, num_channels],
                dtype=np.int_ if return_count else np.bool_,
            ),
            timestamps=time[:-1],
            rate=1.0 / bin_size,
        )
        for idx, spiketrain in enumerate(self.data):
            bins = np.digitize(spiketrain, time)
            bincount = np.bincount(bins, minlength=n_bins + 2)[1:-1]
            if return_count:
                bin_spike = bincount
            else:
                bin_spike = (bincount != 0).astype(np.bool_)
            signal.data[:, idx] = bin_spike
        return signal



#! /usr/bin/env python

# This file is part of the SpikeCoding repository - MAVLab TU Delft
#
# MIT License
#
# Copyright (c) 2021 Julien Dupeyroux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# @author Julien Dupeyroux, Stein Stroobants

import numpy as np
from scipy.stats import norm


def temporal_contrast(data, factor):
    # Based on algorithm provided in:
    #   Sengupta et al. (2017)
    #   Petro et al. (2020)
    diff = np.zeros(len(data) - 1)
    spikes = np.zeros(len(data))
    for i in range(len(data) - 1):
        diff[i] = data[i + 1] - data[i]
    threshold = np.mean(diff) + factor * np.std(diff)
    diff = np.insert(diff, 0, diff[1])
    for i in range(len(data)):
        if diff[i] > threshold:
            spikes[i] = 1
        elif diff[i] < -threshold:
            spikes[i] = -1
    return spikes, threshold


def step_forward(data, threshold):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    startpoint = data[0]
    spikes = np.zeros(len(data))
    base = startpoint
    for i in range(1, len(data)):
        if data[i] > base + threshold:
            spikes[i] = 1
            base = base + threshold
        elif data[i] < base - threshold:
            spikes[i] = -1
            base = base - threshold
    return spikes, startpoint


def moving_window(data, threshold, window):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    startpoint = data[0]
    spikes = np.zeros(len(data))
    base = np.mean(data[0 : window + 1])
    for i in range(window + 1):
        if data[i] > base + threshold:
            spikes[i] = 1
        elif data[i] < base - threshold:
            spikes[i] = -1
    for i in range(window + 2, len(data)):
        base = np.mean(data[(i - window - 1) : (i - 1)])
        if data[i] > base + threshold:
            spikes[i] = 1
        elif data[i] < base - threshold:
            spikes[i] = -1
    return spikes, startpoint


def hough_spike(data, fir):
    # Based on algorithm provided in:
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift * np.ones(len(data))
    for i in range(len(data)):
        count = 0
        for j in range(len(fir)):
            if i + j < len(data):
                if data[i + j] >= fir[j]:
                    count = count + 1
        if count == len(fir):
            spikes[i] = 1
            for j in range(len(fir)):
                if i + j < len(data):
                    data[i + j] = data[i + j] - fir[j]
    return spikes, shift


def modified_hough_spike(data, fir, threshold):
    # Based on algorithm provided in:
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift * np.ones(len(data))
    for i in range(len(data)):
        error = 0
        for j in range(len(fir)):
            if i + j < len(data):
                if data[i + j] < fir[j]:
                    error = error + fir[j] - data[i + j]
        if error <= threshold:
            spikes[i] = 1
            for j in range(len(fir)):
                if i + j < len(data):
                    data[i + j] = data[i + j] - fir[j]
    return spikes, shift


def ben_spike(data, fir, threshold):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    #   Sengupta et al. (2017)
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift * np.ones(len(data))
    for i in range(len(data) - len(fir) + 1):
        err1 = 0
        err2 = 0
        for j in range(len(fir)):
            err1 = err1 + abs(data[i + j] - fir[j])
            err2 = err2 + abs(data[i + j - 1])
        if err1 <= err2 * threshold:
            spikes[i] = 1
            for j in range(len(fir)):
                if i + j + 1 < len(data):
                    data[i + j + 1] = data[i + j + 1] - fir[j]
    return spikes, shift


def grf_spike(data, m, min_input, max_input):
    # Adapted from algorithm provided in:
    #   Bohté et al. (2002)
    # Modifications: definition of sigma, removal of beta constant,
    #                and modified WTA process

    if np.isscalar(data):
        data = [data]

    spikes = np.zeros((len(data), m))
    neuron_outputs = np.zeros(m)

    for j in range(len(data)):
        for i in range(m):
            mu = min_input + (2 * (i + 1) - 3) / 2 * (max_input - min_input) / (m - 2)
            sigma = (max_input - min_input) / (m - 2)
            neuron_outputs[i] = norm.pdf(data[j], mu, sigma)

        spikes[j, np.argmax(neuron_outputs)] = 1
    return spikes


def one_hot_place_spike(data, m, min_input, max_input):
    # Simple population coding algorithm adapted from Stagsted et al. (2020) that represents inputs by a location.
    # An input is assigned to the neuron that is closest to its value.
    # Only one neuron fires at every timestep

    if np.isscalar(data):
        data = [data]

    spikes = np.zeros((len(data), m))

    for j in range(len(data)):
        size_change = (
            1 / 2 * (max_input - min_input) / (m - 2)
        )  # to make sure it has the same lower/upper bounds as the Bohte paper
        idx = int(
            np.round(
                (
                    (data[j] - (min_input - size_change))
                    / ((max_input + size_change) - (min_input - size_change))
                )
                * (m - 1)
            )
        )
        spikes[j, idx] = 1

    return spikes


def grf_spike_with_internal_timesteps(
    data, min_input, max_input, neurons=10, timesteps=10, beta=1.5
):
    """Create a series of spikes based on Gaussian Receptive Fields
    Adapted from algorithm provided in:
        Bohté et al. (2002)

    Keyword arguments:
    data --
    neurons -- numbers of neurons (default 10)
    timesteps -- number of timesteps (default 10)
    min_input -- minimal value
    max_input -- maximum value
    beta -- tuning parameter that determines the width of the receptive fields
    """

    if np.isscalar(data):
        data = [data]

    spikes = np.zeros((len(data), timesteps, neurons))
    responses = np.zeros(neurons)

    # Calculation of mu and sigma of the Gaussian receptive fields
    mu = min_input + (2 * (np.arange(neurons) + 1) - 3) / 2 * (
        max_input - min_input
    ) / (neurons - 2)
    sigma = 1 / beta * (max_input - min_input) / (neurons - 2)
    max_prob = norm.pdf(mu[0], mu[0], sigma)

    for j in range(len(data)):
        for i in range(neurons):
            responses[i] = norm.pdf(data[j], mu[i], sigma)
            size_change = max_prob / (2 * timesteps)
            new = int(
                np.round(
                    (
                        (responses[i] + size_change)
                        / (max_prob + 2 * size_change)
                        * (timesteps + 1)
                    )
                    + 0.0001
                )
            )  # 0.0001 for roundoff errors...
            spiking_time = timesteps - new
            if spiking_time < timesteps - 1:
                spikes[j, spiking_time, i] = 1
    spikes = spikes.reshape([len(data) * timesteps, neurons])
    return spikes



__doc__ = ""
__all__ = ["ButterBandpass"]

from typing import Optional

import os
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.signal as sps

from miv.core.datatype import Signal
from miv.core.operator import OperatorMixin
from miv.core.wrapper import wrap_generator_to_generator
from miv.typing import SignalType


@dataclass
class ButterBandpass(OperatorMixin):
    """Classical bandpass filter using `scipy` butterworth filter

    Parameters
    ----------
    lowcut : float
        low-pass frequency
    highcut : float
        high-pass frequency
    order : int
        The order of the filter. (default=5)
    tag : str
        Tag for the collection of filter.
    btype : str
        Filter type: bandpass, lowpass, highpass, bandstop. (default="bandpass")
        If set to lowpass, the critical frequency is set to 'highcut'.
        If set to highpass, the critical frequency is set to 'lowcut'.
        If set to bandpass or bandstop, the critical frequency is '[lowcut, highcut]'.
    """

    lowcut: Optional[float] = None
    highcut: Optional[float] = None
    order: int = 5
    tag: str = "bandpass filter"
    btype: str = "bandpass"

    @wrap_generator_to_generator
    def __call__(self, signal: Signal) -> Signal:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            signal

        Returns
        -------
        Signal

        """
        rate = signal.rate
        b, a = self._butter_bandpass(rate)
        y = signal.data.copy()
        num_channel = signal.number_of_channels
        for ch in range(num_channel):
            y[:, ch] = sps.lfilter(b, a, signal.data[:, ch])
        return Signal(data=y, timestamps=signal.timestamps, rate=rate)

    def __post_init__(self):
        if self.lowcut is not None and self.highcut is not None:
            assert (
                self.lowcut < self.highcut
            ), f"{self.lowcut=} should be lower than {self.highcut=}."
            assert (
                min(self.lowcut, self.highcut) > 0
            ), "Filter critical frequencies must be greater than 0"
        elif self.lowcut is not None:
            assert self.lowcut > 0, "Filter frequencies must be greater than 0"
        elif self.highcut is not None:
            assert self.highcut > 0, "Filter frequencies must be greater than 0"
        assert self.order > 0 and isinstance(
            self.order, int
        ), f"Filter {self.order} must be an nonnegative integer."
        assert (
            self.lowcut is not None or self.highcut is not None
        ), "Filtering frequency cannot be both None"
        super().__init__()

    def _butter_bandpass(self, sampling_rate: float):
        nyq = 0.5 * sampling_rate
        if self.btype in ["bandpass", "bandstop"]:
            low = self.lowcut / nyq
            high = self.highcut / nyq
            critical_frequency = [low, high]
        elif self.btype == "highpass":
            critical_frequency = self.lowcut / nyq
        elif self.btype == "lowpass":
            critical_frequency = self.highcut / nyq
        else:
            raise ValueError("Unknown btype: %s" % self.btype)
        b, a = sps.butter(self.order, critical_frequency, btype=self.btype)
        return b, a

    # def plot_frequency_response(self, signal:Signal, show=False, save_path=None):
    #     """plot_frequency_response

    #     Parameters
    #     ----------
    #     signal : Signal

    #     Returns
    #     -------
    #     plt.Figure
    #     """
    #     sampling_rate = next(signal).rate
    #     b, a = self._butter_bandpass(sampling_rate)
    #     w, h = sps.freqs(b, a)
    #     fig = plt.figure()
    #     plt.semilogx(w, 20 * np.log10(abs(h)))
    #     plt.title(
    #         f"Butterworth filter (order{self.order}) frequency response [{self.lowcut},{self.highcut}]"
    #     )
    #     plt.xlabel("Frequency")
    #     plt.ylabel("Amplitude")
    #     if show:
    #         plt.show()
    #     if save_path is not None:
    #         fig.savefig(os.path.join(save_path, 'filter_frequency_response.png'))
    #     return fig



from __future__ import annotations

__doc__ = """"""
__all__ = ["_Chainable", "BaseChainingMixin"]

from typing import TypeVar  # TODO: For python 3.11, we can use typing.Self
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    List,
    Optional,
    Protocol,
    Set,
    Union,
)

import functools
import itertools

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from miv.core.datatype import DataTypes

SelfChain = TypeVar("SelfChain", bound="_Chainable")


class _Chainable(Protocol):
    """
    Behavior includes:
        - Chaining modules in forward/backward linked lists
            - Forward direction defines execution order
            - Backward direction defines dependency order
    """

    @property
    def tag(self) -> str:
        ...

    @property
    def output(self) -> list[DataTypes]:
        ...

    def __rshift__(self, right: SelfChain) -> SelfChain:
        ...

    def iterate_upstream(self) -> Iterator[SelfChain]:
        ...

    def iterate_downstream(self) -> Iterator[SelfChain]:
        ...

    def clear_connections(self) -> None:
        ...

    def summarize(self) -> str:
        """Print summary of downstream network structures."""
        ...


class BaseChainingMixin:
    """
    Base mixin to create chaining structure between objects.

    Need further implementation of: output, tag
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._downstream_list: list[_Chainable] = []
        self._upstream_list: list[_Chainable] = []

        self._output: DataTypes | None = None

    def __rshift__(self, right: _Chainable) -> _Chainable:
        self._downstream_list.append(right)
        right._upstream_list.append(self)
        return right

    def clear_connections(self) -> None:
        """Clear all the connections to other nodes, and remove dependencies."""
        for node in self.iterate_downstream():
            node._upstream_list.remove(self)
        for node in self.iterate_upstream():
            node._downstream_list.remove(self)
        self._downstream_list.clear()
        self._upstream_list.clear()

    def iterate_downstream(self) -> Iterator[_Chainable]:
        return iter(self._downstream_list)

    def iterate_upstream(self) -> Iterator[_Chainable]:
        return iter(self._upstream_list)

    def summarize(self) -> str:  # TODO: create DFS and BFS traverser
        q = [(0, self)]
        order = []  # DFS
        while len(q) > 0:
            depth, current = q.pop(0)
            if current in order:  # Avoid loop
                continue
            q += [(depth + 1, node) for node in current.iterate_downstream()]
            order.append((depth, current))
        return self._text_visualize_hierarchy(order)

    def visualize(self, show: bool = False, seed: int = 200) -> None:
        import networkx as nx

        G = nx.DiGraph()

        # BFS
        visited = []
        next_list = [self]
        while next_list:
            v = next_list.pop()
            visited.append(v)
            for node in itertools.chain(v.iterate_downstream()):
                G.add_edge(v.tag, node.tag)
                if node in visited or node in next_list:
                    continue
                visited.append(node)
                next_list.append(node)

        # Draw the graph
        # TODO: Balance layout
        # TODO: On edge, label the argument order
        pos = nx.spring_layout(G, seed=seed)
        nx.draw_networkx_nodes(G, pos, node_size=500, alpha=0.8)
        nx.draw_networkx_edges(G, pos, width=2, arrows=True)
        nx.draw_networkx_labels(G, pos, font_size=12, font_family="sans-serif")

        # Display the graph
        plt.margins(x=0.4)
        plt.axis("off")
        if show:
            plt.show()
        return G

    def _text_visualize_hierarchy(self, string_list, prefix="|__ "):
        output = []
        for i, item in enumerate(string_list):
            depth, label = item
            if depth > 0:
                output.append("    " * (depth - 1) + prefix + str(label))
            else:
                output.append(str(label))
        return "\n".join(output)

    def _get_upstream_topology(self, lst: list[SelfChain] = None) -> list[SelfChain]:
        if lst is None:
            lst = []
        if (
            hasattr(self, "cacher")
            and self.cacher is not None
            and self.cacher.cache_dir is not None
            and self.cacher.check_cached()
        ):
            pass
        else:
            for node in self.iterate_upstream():
                if node in lst:
                    continue
                node._get_upstream_topology(lst)
        lst.append(self)
        return lst

    def topological_sort(self):
        """
        Topological sort of the graph.
        Raise RuntimeError if there is a loop in the graph.
        """
        # TODO: Make it free function
        upstream = self._get_upstream_topology()

        # pos = dict()  # FIXME: Operators are not hashable
        key = []
        pos = []
        ind = 0
        tsort = []

        while len(upstream) > 0:
            key.append(upstream[-1])
            pos.append(ind)
            # pos[upstream[-1]] = ind
            tsort.append(upstream[-1])
            ind += 1
            upstream.pop()
        for source in tsort:
            for up in source.iterate_upstream():
                if up not in key:
                    continue
                before = pos[key.index(source)]
                after = pos[key.index(up)]

                # If parent vertex does not appear first
                if before > after:
                    raise RuntimeError(
                        f"Found loop in operation stream: node {source} is already in the upstream : {up}."
                    )

        tsort.reverse()
        return tsort



__all__ = ["GridMEA"]

from typing import Tuple

import matplotlib
import numpy as np

from miv.mea.base import MEAMixin


class GridMEA(MEAMixin):
    """
    A class representing a grid-based multi-electrode array (MEA).

    Example
    -------
    For example, you could create an instance of the GridMEA class like this:
    literal blocks::

        grid = np.arrange(9).reshape(3, 3)
        mea = GridMEA(grid)
    """

    def __init__(
        self,
        grid: np.ndarray,
        spacing: Tuple[float, float] = (100, 100),
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert grid.ndim == 2, "The grid must be 2-D."
        assert spacing[0] > 0 and spacing[1] > 0, "The spacing must be positive."
        self.grid = grid
        self.spacing = spacing

        self.nrow, self.ncol = grid.shape

    def to_json(self) -> dict:
        """Return a JSON-serializable dictionary"""
        return {
            "grid": self.grid.tolist(),
            "spacing": self.spacing,
        }

    def map_data(
        self, vector: np.ndarray, missing_value: float = 0.0
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        value_grid = np.full_like(self.grid, missing_value)
        for idx, value in enumerate(vector):
            if idx not in self.grid:
                continue
            value_grid[self.grid == idx] = value
        X = np.arange(self.ncol) * self.spacing[0]
        Y = np.arange(self.nrow) * self.spacing[1]
        Xn, Yn = np.meshgrid(X, Y)
        return Xn, Yn, value_grid

    @property
    def coordinates(self):
        """Return the coordinates of the electrodes"""
        X = np.arange(self.ncol) * self.spacing[0]
        Y = np.arange(self.nrow) * self.spacing[1]
        Xn, Yn = np.meshgrid(X, Y)
        return np.vstack([Xn.ravel(), Yn.ravel()]).T



__doc__ = """
ASDF file type loader

author: tvarley
        djhavert

.. autoclass:: miv.io.asdf.asdf.DataASDF
   :members:

"""
__all__ = ["DataASDF"]

import os
import sys

import numpy as np
import pandas as pd
from scipy.io import loadmat

from miv.core.datatype import Spikestamps
from miv.core.operator import DataLoaderMixin


class DataASDF(DataLoaderMixin):
    """ASDF file type loader"""

    def __init__(self, data_path, rate: float):  # pragma: no cover
        """
        Constructor

        :param data_path: path to data file
        :param rate: sampling rate
        """
        self.data_path = data_path
        self.rate = rate

    def load(self):  # pragma: no cover
        """
        Load data from ASDF file
        """
        # load spike time info
        asdf = loadmat(f"{self.data_path}", appendmat=False)["asdf_raw"]
        # info = np.squeeze(asdf[len(asdf) - 1].item())
        # nNeu = info[0]
        # time_end = info[1]

        # create spikestamps
        spikestamps = Spikestamps()
        # for each neuron
        for i in range(asdf.shape[0] - 2):
            # get spike times
            times = np.squeeze(asdf[i].item())
            times = times.astype(int) / self.rate
            # need to modify to fix error, "tuple index out of range"- think it's if something in the asdf is empty, but not sure
            spikestamps.append(times)
        return spikestamps



from miv.visualization.activity import *
from miv.visualization.causality import *
from miv.visualization.connectivity import *
from miv.visualization.event import *
from miv.visualization.fft_domain import *
from miv.visualization.raw_signal import *



__doc__ = ""
__all__ = ["Notch"]

from typing import Optional

import os
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.signal as sps

from miv.core.datatype import Signal
from miv.core.operator import OperatorMixin
from miv.core.wrapper import wrap_generator_to_generator
from miv.typing import SignalType


@dataclass
class Notch(OperatorMixin):
    """Notch filter using `scipy` second-order notch filter, wrapped in operator mixin.

    Parameters
    ----------
    w0 : float
        notch frequency
    Q : float
        Q factor
    tag : str
        Tag for the collection of filter.
    """

    f0: Optional[float] = 60.0  # Hz
    Q: Optional[float] = 30.0
    tag: str = "notch filter"

    @wrap_generator_to_generator
    def __call__(self, signal: Signal) -> Signal:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            signal

        Returns
        -------
        Signal

        """
        rate = signal.rate
        b, a = sps.iirnotch(w0=self.f0, Q=self.Q, fs=rate)
        y = signal.data.copy()
        num_channel = signal.number_of_channels
        for ch in range(num_channel):
            y[:, ch], h = sps.freqz(b, a, signal.data[:, ch])
        return Signal(data=y, timestamps=signal.timestamps, rate=rate)

    def __post_init__(self):
        super().__init__()

    # def plot_frequency_response(self, signal:Signal, show=False, save_path=None):
    #     """plot_frequency_response

    #     Parameters
    #     ----------
    #     signal : Signal

    #     Returns
    #     -------
    #     plt.Figure
    #     """
    #     sampling_rate = next(signal).rate
    #     b, a = self._butter_bandpass(sampling_rate)
    #     w, h = sps.freqs(b, a)
    #     fig = plt.figure()
    #     plt.semilogx(w, 20 * np.log10(abs(h)))
    #     plt.title(
    #         f"Butterworth filter (order{self.order}) frequency response [{self.lowcut},{self.highcut}]"
    #     )
    #     plt.xlabel("Frequency")
    #     plt.ylabel("Amplitude")
    #     if show:
    #         plt.show()
    #     if save_path is not None:
    #         fig.savefig(os.path.join(save_path, 'filter_frequency_response.png'))
    #     return fig



from __future__ import annotations

__doc__ = """"""
__all__ = [
    "Operator",
    "DataLoader",
    "DataLoaderMixin",
    "OperatorMixin",
    "DataNodeMixin",
    "DataNode",
]

from typing import TYPE_CHECKING, Callable, Generator, List, Optional, Protocol, Union

import functools
import itertools
import os
import pathlib
from dataclasses import dataclass

if TYPE_CHECKING:
    from miv.core.datatype import DataTypes

from miv.core.operator.cachable import (
    DataclassCacher,
    FunctionalCacher,
    _Cachable,
    _CacherProtocol,
)
from miv.core.operator.callback import BaseCallbackMixin, _Callback
from miv.core.operator.chainable import BaseChainingMixin, _Chainable
from miv.core.policy import VanillaRunner, _Runnable, _RunnerProtocol


class Operator(
    _Callback,
    _Chainable,
    _Cachable,
    _Runnable,
    Protocol,
):
    """ """

    def run(self, dry_run: bool = False) -> None:
        ...


class DataLoader(
    _Callback,
    _Chainable,
    _Cachable,
    _Runnable,
    Protocol,
):
    """ """

    def load(self) -> Generator[DataTypes, None, None]:
        ...


class DataNode(_Chainable, _Runnable, Protocol):
    ...


class DataNodeMixin(BaseChainingMixin):
    """ """

    def __init__(self):
        super().__init__()
        self._output = None

    @property
    def output(self) -> list[DataTypes]:
        self._output = self
        return self._output

    def run(self, **kwargs):
        pass


class DataLoaderMixin(BaseChainingMixin, BaseCallbackMixin):
    """ """

    def __init__(self):
        super().__init__()
        self._output: DataTypes | None = None

        self.runner = VanillaRunner()
        self.cacher = FunctionalCacher(self)
        self.cacher.cache_dir = os.path.join(self.data_path, ".cache")

    @property
    def output(self) -> list[DataTypes]:
        self._output = self.load()
        return self._output

    def run(self, **kwargs):
        pass


class OperatorMixin(BaseChainingMixin, BaseCallbackMixin):
    """
    Behavior includes:
        - Whenever "run()" method is executed:
            1. Check if the module is cached in the same parameters. Y: Pass to 6
            2. Callback: before run
            3. Run
            4. Callback: after run
            5. Save cache
            6. Exit
        - Cache includes:
            - All results from callback
    """

    def __init__(self):
        super().__init__()
        self._output: DataTypes | None = None
        self.runner = VanillaRunner()
        self.cacher = DataclassCacher(self)

        assert self.tag != ""
        self.set_save_path("results")  # Default analysis path

    def set_caching_policy(self, cacher: _CacherProtocol):
        self.cacher = cacher(self)

    def set_save_path(self, path: str | pathlib.Path):
        self.analysis_path = os.path.join(path, self.tag.replace(" ", "_"))
        self.cacher.cache_dir = os.path.join(self.analysis_path, ".cache")

    def receive(self) -> list[DataTypes]:
        return [node.output for node in self.iterate_upstream()]

    @property
    def output(self) -> list[DataTypes]:
        self._execute()
        return self._output  # TODO: Just use upstream caller instead of .output

    def _execute(self):
        args: list[DataTypes] = self.receive()  # Receive data from upstream
        # TODO : implement pre-run callback
        # args = self.callback_before_run(args)
        if len(args) == 0:
            output = self.runner(self.__call__)
        else:
            output = self.runner(self.__call__, args)
        # Post Processing
        self._output = self.callback_after_run(output)

    def run(
        self,
        save_path: str | pathlib.Path | None = None,
        dry_run: bool = False,
        skip_plot: bool = False,
    ) -> None:
        # Execute the module
        if save_path is not None:
            self.set_save_path(save_path)

        if dry_run:
            print("Dry run: ", self.__class__.__name__)
            return
        self._execute()
        if not skip_plot:
            self.plot(show=False, save_path=True, dry_run=dry_run)



__doc__ = """
Signal
======

.. autoclass:: Signal
   :members:

"""

__all__ = ["Signal"]

from typing import Optional, Tuple

from dataclasses import dataclass

import numpy as np

from miv.core.operator.operator import DataNodeMixin
from miv.core.policy import SupportMultiprocessing
from miv.typing import SignalType, TimestampsType


@dataclass
class Signal(SupportMultiprocessing, DataNodeMixin):
    """
    Contiguous array of raw signal type.

    [signal length, number of channels]
    """

    _CHANNELAXIS = 1
    _SIGNALAXIS = 0

    data: SignalType
    timestamps: TimestampsType
    rate: int = 30_000

    def __post_init__(self):
        self.data = np.asarray(self.data)
        assert len(self.data.shape) == 2, "Signal must be 2D array"

    @property
    def number_of_channels(self) -> int:
        """Number of channels in the signal."""
        return self.data.shape[self._CHANNELAXIS]

    def __getitem__(self, i: int) -> SignalType:
        return self.data[:, i]  # TODO: Fix to row-major

    def select(self, indices: Tuple[int, ...]) -> "Signal":
        """Select channels by indices."""
        return Signal(self.data[:, indices], self.timestamps, self.rate)

    def get_start_time(self):
        """Get the start time of the signal."""
        return self.timestamps.min()

    def get_end_time(self):
        """Get the end time of the signal."""
        return self.timestamps.max()

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the signal."""
        return self.data.shape

    def append(self, value) -> None:
        """Append a channels to the end of the existing signal."""
        assert value.shape[self._SIGNALAXIS] == self.data.shape[self._SIGNALAXIS]
        self.data = np.append(self.data, value, axis=self._CHANNELAXIS)

    def extend_signal(self, value: SignalType, time: TimestampsType) -> None:
        """Append a signal to the end of the existing signal."""
        assert value.shape[self._SIGNALAXIS] == time.shape[0]
        assert (
            value.shape[self._CHANNELAXIS] == self.data.shape[self._CHANNELAXIS]
        ), "Signal must have same number of channels"
        self.data = np.append(self.data, value, axis=self._SIGNALAXIS)
        self.timestamps = np.append(self.timestamps, time)

    def prepend_signal(self, value: SignalType, time: TimestampsType) -> None:
        """Prepend a signal to the end of the existing signal."""
        assert (
            value.shape[self._SIGNALAXIS] == time.shape[0]
        ), "Time and signal must have same length"
        assert (
            value.shape[self._CHANNELAXIS] == self.data.shape[self._CHANNELAXIS]
        ), "Signal must have same number of channels"
        self.data = np.append(value, self.data, axis=self._SIGNALAXIS)
        self.timestamps = np.append(self.timestamps, time)



__all__ = ["MEAGeometryProtocol"]

import typing
from typing import Any, Iterable, Protocol, Tuple

import matplotlib
import numpy as np

from miv.core.operator.cachable import _Jsonable
from miv.core.operator.chainable import _Chainable
from miv.core.policy import _Runnable


class MEAGeometryProtocol(_Jsonable, _Chainable, _Runnable, Protocol):
    @property
    def coordinates(self) -> np.ndarray:
        """Return coordinates of MEA electrodes location"""
        ...

    def get_xy(self, idx: int) -> Tuple[float, float]:
        """Given node index, return xy coordinate"""
        ...

    def save(self, path: str) -> None:
        ...

    def load(self, path: str) -> None:
        ...

    def view(self) -> matplotlib.pyplot.Figure:
        """Simplified view of MEA orientation"""
        ...

    def map_data(self, vector: np.ndarray, missing_value: float) -> np.ndarray:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        ...



__doc__ = """OpenEphys sample data."""
__all__ = ["load_data"]

import gzip
import os

import numpy as np

from miv.datasets.utils import get_file
from miv.io.openephys import DataManager


def load_data(progbar_disable: bool = False):  # pragma: no cover
    """
    Loads the sample recorded data from OpenEphys aquisition system. `Direct Download <https://uofi.box.com/shared/static/daagmwebzfdl1jjgfhjzzhof2if44m04.zip>`_

    Total size: 21.2 kB (compressed)

    File hash: ab1b282a2560e75263188dc2bab56b019bb83ea12aa63692c9560c6e8281de29

    Examples
    --------
        >>> from miv.datasets.ttl_events import load_data
        >>> experiments: miv.io.DataManager = load_data()
        datasets/ttl_recording/sample_event_recording
            0: <miv.io.data.Data object at 0x7fec71c99fa0>
               └── Record Node 101/experiment1/recording1


    Notes
    -----
    All experiment are 1 minute long, 30k Hz recording of optogenetic
    neuron cells over 64 channels MEA. Dataset includes 1 spontaneous
    recording.

    Spontaneous recording is the recording over 10 seconds period.

    Containing experiments:

    * experiment0: 10 seconds spontaneous recording

    Returns
    -------
    dataset: miv.io.DataManager

    Examples
    --------
        >>> from miv import datasets
        >>> experiments: miv.io.DataManager = datasets.openephys_sample.load_data()

    """

    subdir = "OpenEphys"
    base_url = "https://uofi.box.com/shared/static/daagmwebzfdl1jjgfhjzzhof2if44m04.zip"
    file = "OpenEphys_Sample_64MEA.zip"
    file_hash = "ab1b282a2560e75263188dc2bab56b019bb83ea12aa63692c9560c6e8281de29"

    path = get_file(
        file_url=base_url,
        directory=subdir,
        fname=file,
        file_hash=file_hash,
        archive_format="zip",
        progbar_disable=progbar_disable,
    )
    experiment = DataManager(path)
    experiment.tree()
    return experiment



__all__ = ["NeuralActivity"]

import logging
import os
from dataclasses import dataclass

import matplotlib.animation as manimation
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.policy import StrictMPIRunner
from miv.mea import MEAGeometryProtocol
from miv.statistics import spike_counts_with_kernel
from miv.visualization.utils import interp_2d


@dataclass
class NeuralActivity(OperatorMixin):
    bin_size: float = 0.001  # sec
    firing_rate_interval: float = 1.0  # sec
    skip_interval: int = 50  # frames to skip. TODO: refactor

    tag: str = "neural activity render"

    progress_bar: bool = False

    fps: int = 25
    minimum_split_size: int = 60  # sec. Each video is at least 1 min

    def __post_init__(self):
        super().__init__()
        self.runner = StrictMPIRunner()

    def __call__(self, spikestamps: Spikestamps, mea: MEAGeometryProtocol):
        comm = self.runner.comm
        rank = self.runner.get_rank()
        size = self.runner.get_size()

        # Binning in first rank
        if self.runner.is_root():
            spiketrains_bins = spikestamps.binning(self.bin_size, return_count=True)
            probe_times = spiketrains_bins.timestamps[:: self.skip_interval]
        else:
            spiketrains_bins = None
            probe_times = None
        spiketrains_bins = comm.bcast(spiketrains_bins, root=self.runner.get_root())
        probe_times = comm.bcast(probe_times, root=self.runner.get_root())

        # Split Tasks
        num_frames = probe_times.shape[0]
        if size * self.fps * self.minimum_split_size > num_frames:
            size = np.ceil(num_frames / (self.fps * self.minimum_split_size)).astype(
                np.int_
            )
            if rank == self.runner.get_root():
                logging.warning(
                    f"Too many ranks. Splitting into {size} tasks for total {num_frames} frames."
                )
            if rank >= size:
                return

        probe_times = np.array_split(probe_times, size)[rank]
        start_time = probe_times[0]
        end_time = probe_times[-1]
        logging.info(
            f"{rank=} | rendering from {start_time=:.03f} to {end_time=:.03f}: {probe_times.shape[0]} frames."
        )

        xs = []
        for i in range(spiketrains_bins.number_of_channels):
            x = spike_counts_with_kernel(
                spikestamps[i],
                probe_times,
                lambda x: np.logical_and(x > 0, x < self.firing_rate_interval).astype(
                    np.float_
                ),
            )
            xs.append(x)
        xs = np.asarray(xs)

        xmax = np.max(xs)
        xmin = np.min(xs)

        # Output Images
        FFMpegWriter = manimation.writers["ffmpeg"]
        metadata = dict(
            title="Movie Test", artist="Matplotlib", comment="Movie support!"
        )
        writer = FFMpegWriter(fps=self.fps, metadata=metadata)

        os.makedirs(self.analysis_path, exist_ok=True)
        video_name = os.path.join(
            self.analysis_path, f"output_from_{start_time:.03f}_to_{end_time:.03f}.mp4"
        )

        fig = plt.figure(figsize=(8, 6))
        with writer.saving(fig, video_name, dpi=200):  # TODO: cut each video into 1min
            for timestep, time in tqdm(
                enumerate(probe_times),
                desc=f"Rendering {rank}/{size}",
                position=rank,
                total=probe_times.shape[0],
                disable=not self.progress_bar,
            ):
                X, Y, Z = mea.map_data(xs[:, timestep])

                fig.clf()
                ax = fig.add_subplot(111)
                # X, Y, Z = interp_2d(Z)
                ax.plot(mea.coordinates[:, 0], mea.coordinates[:, 1], "k.", ms=1)
                pcm = ax.pcolormesh(
                    X, Y, Z, cmap="Oranges", vmin=xmin, vmax=xmax, shading="gouraud"
                )
                cbar = fig.colorbar(pcm, ax=ax)
                cbar.ax.set_ylabel(
                    f"activity per {self.firing_rate_interval:.03f} sec", rotation=270
                )

                ax.set_aspect("equal")
                ax.set_xlabel("channels x-axis")
                ax.set_ylabel("channels y-axis")
                ax.set_title(f"Spatial Neural Activity ({time:.03f} sec)")

                writer.grab_frame()
        plt.close(plt.gcf())

        return



__doc__ = """
Events
======

.. autoclass:: Events
   :members:

"""

__all__ = ["Events"]

from typing import List, Optional

from collections import UserList

import numpy as np
import quantities as pq

from miv.core.datatype.collapsable import CollapseExtendableMixin
from miv.core.datatype.signal import Signal
from miv.core.operator.operator import DataNodeMixin


class Events(CollapseExtendableMixin, DataNodeMixin):
    """
    A list of events.

    Comply with `Extendable` protocols.
    """

    def __init__(self, data: List[float] = None):
        super().__init__()
        self.data = data if data is not None else []

    def append(self, item):
        raise NotImplementedError("Not implemented yet. Need to append and sort")

    def extend(self, other):
        raise NotImplementedError("Not implemented yet. Need to extend and sort")

    def __len__(self):
        return len(self.data)

    def get_last_event(self):
        """Return timestamps of the last event"""
        return max(self.data)

    def get_first_event(self):
        """Return timestamps of the first event"""
        return min(self.data)

    def get_view(self, t_start: float, t_end: float):
        """Truncate array and only includes spikestamps between t_start and t_end."""
        return Events(sorted(list(filter(lambda x: t_start <= x <= t_end, self.data))))

    def binning(
        self,
        bin_size: float = 1 * pq.ms,
        t_start: Optional[float] = None,
        t_end: Optional[float] = None,
        return_count: bool = False,
    ) -> Signal:
        """
        Forms a binned events

        Parameters
        ----------
        bin_size : float | pq.Quantity
            bin size in the unit of time.
        return_count : bool
            If set to true, return the bin count. (default=False)
        """

        if isinstance(bin_size, pq.Quantity):
            bin_size = bin_size.rescale(pq.s).magnitude
        assert bin_size > 0, "bin size should be greater than 0"

        t_start = self.get_first_event() if t_start is None else t_start
        t_end = self.get_last_event() if t_end is None else t_end
        n_bins = int(np.ceil((t_end - t_start) / bin_size))
        time = t_start + (np.arange(n_bins + 1) * bin_size)

        signal = Signal(
            data=np.zeros(
                [time.shape[0] - 1, 1],
                dtype=np.int_ if return_count else np.bool_,
            ),
            timestamps=time[:-1],
            rate=1.0 / bin_size,
        )

        bins = np.digitize(self.data, time)
        bincount = np.bincount(bins, minlength=n_bins + 2)[1:-1]
        if return_count:
            bin_events = bincount
        else:
            bin_events = (bincount != 0).astype(np.bool_)
        signal.data[:, 0] = bin_events

        return signal



__doc__ = """

Code Example::

    detection = ThresholdCutoff(cutoff=3.5)
    spiketrains = detection(signal, timestamps, sampling_rate)

.. currentmodule:: miv.signal.spike

.. autosummary::
   :nosignatures:
   :toctree: _toctree/DetectionAPI

   SpikeDetectionProtocol
   ThresholdCutoff

"""
__all__ = ["ThresholdCutoff", "query_firing_rate_between"]

from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

import csv
import functools
import inspect
import logging
import multiprocessing
import os
import pathlib
from dataclasses import dataclass

import matplotlib.pyplot as plt
import neo
import numpy as np
import quantities as pq
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.policy import InternallyMultiprocessing
from miv.core.wrapper import wrap_cacher
from miv.statistics import firing_rates
from miv.typing import SignalType, SpikestampsType, TimestampsType


@dataclass
class ThresholdCutoff(OperatorMixin):
    """ThresholdCutoff
    Spike sorting step by step guide is well documented `here <http://www.scholarpedia.org/article/Spike_sorting>`_.

        Attributes
        ----------
        dead_time : float
            (default=0.003)
        search_range : float
            (default=0.002)
        cutoff : Union[float, np.ndarray]
            (default=5.0)
        use_mad : bool
            (default=False)
        tag : str
        units : Union[str, pq.UnitTime]
            (default='sec')
        progress_bar : bool
            Toggle progress bar (default=True)
        return_neotype : bool
            If true, return spiketrains in neo.Spiketrains (default=True)
            If false, return list of numpy-type spiketrains.
    """

    dead_time: float = 0.003
    search_range: float = 0.002
    cutoff: float = 5.0
    use_mad: bool = True
    tag: str = "spike detection"
    progress_bar: bool = False
    units: str = "sec"
    return_neotype: bool = False  # TODO: Remove, shift to spikestamps datatype

    num_proc: int = 1

    # @wrap_generator_to_generator
    @wrap_cacher(cache_tag="spikestamps")
    def __call__(self, signal: SignalType) -> SpikestampsType:
        """Execute threshold-cutoff method and return spike stamps

        Parameters
        ----------
        signal : Signal

        Returns
        -------
        spiketrain_list : List[SpikestampsType]

        """
        if not inspect.isgenerator(
            signal
        ):  # TODO: Refactor in multiprocessing-enabling decorator
            return self._detection(signal)
        else:
            collapsed_result = Spikestamps()
            # with multiprocessing.Pool(self.num_proc) as pool:
            #    #for result in pool.map(functools.partial(ThresholdCutoff._detection, self=self), signal):
            #    inputs = list(signal)
            #    print(inputs)
            #    for result in pool.map(self._detection, inputs): # TODO: Something is not correct here. Check memory usage.
            #        collapsed_result.extend(spiketrain)
            for sig in signal:  # TODO: mp
                collapsed_result.extend(self._detection(sig))
            return collapsed_result

    # @staticmethod
    def _detection(self, signal: SignalType):
        # Spike detection for each channel
        spiketrain_list = []
        num_channels = signal.number_of_channels  # type: ignore
        timestamps = signal.timestamps
        rate = signal.rate
        for channel in tqdm(range(num_channels), disable=not self.progress_bar):
            array = signal[channel]  # type: ignore

            # Spike Detection: get spikestamp
            spike_threshold = self._compute_spike_threshold(
                array, cutoff=self.cutoff, use_mad=self.use_mad
            )
            crossings = self._detect_threshold_crossings(
                array, rate, spike_threshold, self.dead_time
            )
            spikes = self._align_to_minimum(array, rate, crossings, self.search_range)
            spikestamp = spikes / rate + timestamps.min()
            # Convert spikestamp to neo.SpikeTrain (for plotting)
            if self.return_neotype:
                spiketrain = neo.SpikeTrain(
                    spikestamp,
                    units=self.units,  # TODO: make this compatible to other units
                    t_stop=timestamps.max(),
                    t_start=timestamps.min(),
                )
                spiketrain_list.append(spiketrain)
            else:
                spiketrain_list.append(spikestamp.astype(np.float_))
        spikestamps = Spikestamps(spiketrain_list)
        return spikestamps

    def __post_init__(self):
        super().__init__()

    def _compute_spike_threshold(
        self, signal: SignalType, cutoff: float = 5.0, use_mad: bool = True
    ) -> (
        float
    ):  # TODO: make this function compatible to array of cutoffs (for each channel)
        """
        Returns the threshold for the spike detection given an array of signal.

        Spike sorting step by step, `step 2 <http://www.scholarpedia.org/article/Spike_sorting>`_.

        Parameters
        ----------
        signal : np.array
            The signal as a 1-dimensional numpy array
        cutoff : float
            The spike-cutoff multiplier. (default=5.0)
        use_mad : bool
            Noise estimation method. If set to false, use standard deviation for estimation. (default=True)
        """
        if use_mad:
            noise_mid = np.median(np.absolute(signal)) / 0.6745
        else:
            noise_mid = np.std(signal)
        spike_threshold = -cutoff * noise_mid
        return spike_threshold

    def _detect_threshold_crossings(
        self, signal: SignalType, fs: float, threshold: float, dead_time: float
    ):
        """
        Detect threshold crossings in a signal with dead time and return them as an array

        The signal transitions from a sample above the threshold to a sample below the threshold for a detection and
        the last detection has to be more than dead_time apart from the current one.

        Parameters
        ----------
        signal : SignalType
            The signal as a 1-dimensional numpy array
        fs : float
            The sampling frequency in Hz
        threshold : float
            The threshold for the signal
        dead_time : float
            The dead time in seconds.
        """
        dead_time_idx = dead_time * fs
        threshold_crossings = np.diff((signal <= threshold).astype(int) > 0).nonzero()[
            0
        ]
        distance_sufficient = np.insert(
            np.diff(threshold_crossings) >= dead_time_idx, 0, True
        )
        while not np.all(distance_sufficient):
            # repeatedly remove all threshold crossings that violate the dead_time
            threshold_crossings = threshold_crossings[distance_sufficient]
            distance_sufficient = np.insert(
                np.diff(threshold_crossings) >= dead_time_idx, 0, True
            )
        return threshold_crossings

    def _get_next_minimum(self, signal, index, max_samples_to_search):
        """
        Returns the index of the next minimum in the signal after an index

        :param signal: The signal as a 1-dimensional numpy array
        :param index: The scalar index
        :param max_samples_to_search: The number of samples to search for a minimum after the index
        """
        search_end_idx = min(index + max_samples_to_search, signal.shape[0])
        min_idx = np.argmin(signal[index:search_end_idx])
        return index + min_idx

    def _align_to_minimum(self, signal, fs, threshold_crossings, search_range):
        """
        Returns the index of the next negative spike peak for all threshold crossings

        :param signal: The signal as a 1-dimensional numpy array
        :param fs: The sampling frequency in Hz
        :param threshold_crossings: The array of indices where the signal crossed the detection threshold
        :param search_range: The maximum duration in seconds to search for the minimum after each crossing
        """
        search_end = int(search_range * fs)
        aligned_spikes = [
            self._get_next_minimum(signal, t, search_end) for t in threshold_crossings
        ]
        return np.array(aligned_spikes)

    def plot_spiketrain(
        self,
        spikestamps,
        show: bool = False,
        save_path: Optional[pathlib.Path] = None,
    ) -> plt.Axes:
        """
        Plot spike train in raster
        """
        t0 = spikestamps.get_first_spikestamp()
        tf = spikestamps.get_last_spikestamp()

        # TODO: REFACTOR. Make single plot, and change xlim
        term = 60
        n_terms = int(np.ceil((tf - t0) / term))
        if n_terms == 0:
            # TODO: Warning message
            return None
        for idx in range(n_terms):
            spikes = spikestamps.get_view(
                idx * term + t0, min((idx + 1) * term + t0, tf)
            )
            fig, ax = plt.subplots(figsize=(16, 6))
            ax.eventplot(spikes)
            ax.set_xlabel("Time (s)")
            ax.set_ylabel("Channel")
            if save_path is not None:
                plt.savefig(os.path.join(save_path, f"spiketrain_raster_{idx:03d}.png"))
            if not show:
                plt.close("all")
        if show:
            plt.show()
            plt.close("all")
        return ax

    def plot_firing_rate_histogram(self, spikestamps, show=False, save_path=None):
        """Plot firing rate histogram"""
        threshold = 3

        rates = firing_rates(spikestamps)["rates"]
        hist, bins = np.histogram(rates, bins=20)
        logbins = np.logspace(
            np.log10(max(bins[0], 1e-3)), np.log10(bins[-1]), len(bins)
        )
        fig = plt.figure()
        ax = plt.gca()
        ax.hist(rates, bins=logbins)
        ax.axvline(
            np.mean(rates),
            color="r",
            linestyle="dashed",
            linewidth=1,
            label=f"Mean {np.mean(rates):.2f} Hz",
        )
        ax.axvline(
            threshold,
            color="g",
            linestyle="dashed",
            linewidth=1,
            label="Quality Threshold",
        )
        ax.set_xscale("log")
        xlim = ax.get_xlim()
        ax.set_xlabel("Firing rate (Hz) (log-scale)")
        ax.set_ylabel("Count")
        ax.set_xlim([min(xlim[0], 1e-1), max(1e2, xlim[1])])
        ax.legend()
        if save_path is not None:
            fig.savefig(os.path.join(f"{save_path}", "firing_rate_histogram.png"))
            with open(
                os.path.join(f"{save_path}", "firing_rate_histogram.csv"), "w"
            ) as f:
                writer = csv.writer(f)
                writer.writerow(["channel", "firing_rate_hz"])
                data = list(enumerate(rates))
                data.sort(reverse=True, key=lambda x: x[1])
                for ch, rate in data:
                    writer.writerow([ch, rate])
        if show:
            plt.show()

        return ax


def query_firing_rate_between(
    spikestamps: Spikestamps,
    min_firing_rate: float,
    max_firing_rate: float,
) -> np.ndarray:
    """
    Mask channels with firing rates between min_firing_rate and max_firing_rate

    Parameters
    ----------
    spikestamps : Spikestamps
        Spikestamps
    min_firing_rate : float
        Minimum firing rate
    max_firing_rate : float
        Maximum firing rate

    Returns
    -------
    Spikestamps
        Mask of channels with firing rates between min_firing_rate and max_firing_rate
    """
    rates = np.array(firing_rates(spikestamps)["rates"])
    masks = np.logical_and(rates >= min_firing_rate, rates <= max_firing_rate)
    return np.where(masks)[0]



__doc__ = """
Avalanche Analysis
"""

__all__ = ["AvalancheDetection", "AvalancheAnalysis"]

from typing import List, Optional, Tuple, Union

import logging
import os
import sys
from collections import defaultdict
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import quantities as pq
from scipy.optimize import curve_fit
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.wrapper import wrap_cacher


@dataclass
class AvalancheDetection(OperatorMixin):
    r"""
    An operator for finding avalanches in spike trains.
    It returns the size, duration, and avalanches themselves, defined as all the times each neuron spiked in a given avalanche.

    Parameters
    ----------
    bin_size : float
        The size of the bins in seconds. Default is 0.004 seconds.
    threshold : float
        The threshold for the avalanche size. Default is None.
    time_difference : float
        Minimum time difference between avalanches in seconds.
        Default is equal to the bin_size.
    allow_multiple_spike_per_bin : bool
        If True, allow multiple spikes per bin. Default is False.
        This influence the threshold of the avalanche size.
    minimum_bins_in_avalanche : int
        The minimum number of bins in an avalanche. Default is 10.

    Returns
    -------
    starts : np.ndarray
        The start timestamp of each avalanche.
    ends : np.ndarray
        The end timestamp of each avalanche.
    bincount: Signal
        The binned spike count.

        .. math::
            \sum(thisShape[1:] / thisShape[:-1]) / Avalanche.duration
    """

    bin_size: float = 0.002  # in seconds

    # Miscellaneous configurations
    threshold: Optional[float] = None
    time_difference: Optional[float] = None
    allow_multiple_spike_per_bin: bool = False
    minimum_bins_in_avalanche: int = 10

    tag: str = "Avalanche and Criticality Analysis"

    @wrap_cacher("avalanche_detection")
    def __call__(self, spikestamps: Spikestamps):
        bincount = spikestamps.binning(
            bin_size=self.bin_size, return_count=self.allow_multiple_spike_per_bin
        )
        population_firing = bincount.data.sum(
            axis=bincount._CHANNELAXIS
        )  # Spike count accross channel per bin
        if self.threshold is None:
            self.threshold = (
                population_firing[np.nonzero(population_firing)].mean() / 2.0
            )
        # TODO: try to reuse the code from miv.statistics.burst.burst
        events = (population_firing > self.threshold).astype(np.bool_)

        # pad and find where avalanches start or end based on where 0s change to
        # 1s and vice versa
        concat_events = np.concatenate([[False], events, [False]])
        diff_events = concat_events[1:].astype(np.int_) - concat_events[:-1].astype(
            np.int_
        )
        starts = np.where(diff_events == 1)[0]
        ends = np.where(diff_events == -1)[0]

        # remove avalanches that are too short
        if self.minimum_bins_in_avalanche > 1:
            inds = np.where(ends - starts >= self.minimum_bins_in_avalanche)[0]
            starts = starts[inds]
            ends = ends[inds]

        if not np.isclose(self.bin_size, self.time_difference):
            starts2 = starts[1:]
            ends2 = ends[:-1]
            avalanche_interval = (starts2 - ends2) * self.bin_size
            inds = np.arange(len(avalanche_interval))
            inds = inds[avalanche_interval > self.time_difference]

            starts = starts[np.concatenate([[0], inds + 1])]
            ends = ends[np.concatenate([inds, [len(ends) - 1]])]

        return starts, ends, bincount

    def __post_init__(self):
        super().__init__()
        if isinstance(self.bin_size, pq.Quantity):
            self.bin_size = self.bin_size.rescale("s").magnitude
        if isinstance(self.time_difference, pq.Quantity):
            self.time_difference = self.time_difference.rescale("s").magnitude
        if self.time_difference is None:
            self.time_difference = self.bin_size

    def plot_avalanche_on_raster(self, outputs, show=False, save_path=None):
        """Plot firing rate histogram"""
        if show:
            logging.warning(
                "Plotting avalanche on raster is not supported to be shown interactively."
            )
        spikestamps = self.receive()[0]
        starts, ends, bincount = outputs
        starts_time = starts * self.bin_size + bincount.timestamps[0]
        ends_time = ends * self.bin_size + bincount.timestamps[0]

        fig, ax = plt.subplots(figsize=(16, 6))
        ax.eventplot(spikestamps)

        for start, end in zip(starts_time, ends_time):
            ax.axvspan(start, end, facecolor="r", alpha=0.5)

        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Channel")

        # Save
        left, right = ax.get_xlim()
        interval = 30  # sec
        for i in range(int(np.ceil((right - left) / interval))):
            ax.set_xlim(left + i * interval, left + (i + 1) * interval)
            if save_path is not None:
                plt.savefig(
                    os.path.join(save_path, f"raster_avalanche_overlay_{i}.png")
                )
        plt.close(fig)


@dataclass
class AvalancheAnalysis(OperatorMixin):
    """
    Avalanche and Criticality Analysis

    .. example::

        >>> from miv.statistics.criticality import AvalancheDetection, AvalancheAnalysis

        >>> # Avalanche detection
        >>> avalanche_detection = AvalancheDetection()
        >>> avalanche_analysis = AvalancheAnalysis()
        >>> avalanche_detection >> avalanche_analysis

    Parameters
    ----------
    tag: str
        The tag for the operator. Default is "Avalanche and Criticality Analysis".
    """

    tag: str = (
        "Avalanche and Criticality Analysis"  # keep this same as AvalancheDetection
    )

    progress_bar: bool = False

    @wrap_cacher("shape_analysis")
    def __call__(self, inputs):
        starts, ends, bincount = inputs
        bin_size = 1.0 / bincount.rate
        durations = (ends - starts) * bin_size

        # Compute size and branching ratio
        size = np.zeros(starts.size, dtype=np.int_)
        branching_ratio = np.zeros(starts.size, dtype=np.float_)
        avalanches = []
        for idx, (s, e) in tqdm(
            enumerate(zip(starts, ends)),
            total=len(starts),
            desc="Avalanche analysis",
            disable=not self.progress_bar,
        ):
            if s == e:
                raise ValueError("Avalanche size is 0")
            avalanche = bincount.data[s:e, :]
            avalanches.append(avalanche)
            size[idx] = np.count_nonzero(avalanche.sum(axis=0))
            shape = avalanche.sum(axis=1)
            if np.any(np.isclose(shape[:-1], 0)) or np.isclose(size[idx], 0):
                branching_ratio[idx] = 0.0
            else:
                branching_ratio[idx] = np.sum(shape[1:] / shape[:-1]) / size[idx]

        return durations, size, branching_ratio, avalanches, bin_size

    def __post_init__(self):
        super().__init__()

    def plot_power_law_fitting(self, outputs, show=False, save_path=None):
        durations, size, branching_ratio, _, _ = outputs

        def power(x, a, c):
            return c * (x**a)

        def neg_power(x, a, c):
            return c * (x ** (-a))

        nbins = 100

        fig, axes = plt.subplots(1, 3, figsize=(18, 6))
        hist, bins = np.histogram(size, bins=nbins)
        logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        axes[0].hist(size, bins=logbins, histtype="step", label="data")
        axes[0].set_xscale("log")
        axes[0].set_yscale("log")
        axes[0].set_xlabel("size (# channels)")
        axes[0].set_ylabel("Event Frequency")
        hist, bins = np.histogram(size, bins=logbins)
        try:
            popt, pcov = curve_fit(neg_power, bins[:-1][hist > 1], hist[hist > 1])
            tau = popt[0]
            axes[0].plot(logbins, neg_power(logbins, *popt), label=f"fit {tau=:.2f}")
        except RuntimeError:
            tau = 0
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        axes[0].legend()
        axes[0].set_ylim(bottom=5e-1)

        hist, bins = np.histogram(durations, bins=nbins)
        logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        axes[1].hist(durations, bins=logbins, histtype="step", label="data")
        axes[1].set_xscale("log")
        axes[1].set_yscale("log")
        axes[1].set_xlabel("duration (s)")
        axes[1].set_ylabel("Event Frequency")
        hist, bins = np.histogram(durations, bins=logbins)
        try:
            popt, pcov = curve_fit(neg_power, bins[:-1][hist > 1], hist[hist > 1])
            alpha = popt[0]
            axes[1].plot(logbins, neg_power(logbins, *popt), label=f"fit {alpha=:.2f}")
        except RuntimeError:
            alpha = 0
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        axes[1].legend()
        axes[1].set_ylim(bottom=5e-1)

        def width(p, w):
            return 10 ** (np.log10(p) + w / 2.0) - 10 ** (np.log10(p) - w / 2.0)

        values = []
        avearges = []
        for value in np.unique(durations):
            positions = np.array([value])
            axes[2].boxplot(
                size[durations == value],
                positions=positions,
                widths=width(positions, 0.1),
                showfliers=False,
            )
            values.append(value)
            avearges.append(np.mean(size[durations == value]))
        axes[2].set_xscale("log")
        axes[2].set_yscale("log")
        axes[2].set_xlabel("duration (s)")
        axes[2].set_ylabel("Average size (# channels)")
        try:
            popt, pcov = curve_fit(power, values, avearges)
            self.svz = popt[0]
            axes[2].plot(
                logbins, power(logbins, *popt), label=f"fit 1/svz={popt[0]:.2f}"
            )
        except RuntimeError:
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        axes[2].legend()
        axes[2].set_title(f"({(alpha-1)/(tau-1)=:.2f})")

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "avalanche_power_fitting.png"))
        if show:
            plt.show()
        plt.close(fig)

    def plot_branching_ratio_histogram(self, outputs, show=False, save_path=None):
        durations, size, branching_ratio, _, _ = outputs

        nbins = 100

        fig, ax = plt.subplots(1, 1)
        # hist, bins = np.histogram(size, bins=nbins)
        # logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        ax.hist(
            branching_ratio[np.nonzero(branching_ratio)],
            bins=nbins,
            histtype="step",
            label="data",
        )
        ax.set_xlabel("branching ratio")
        ax.set_ylabel("Event Frequency")
        ax.set_title("branching ratio")
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "branching_ratio.png"))
        if show:
            plt.show()
        plt.close(fig)

    def plot_avalanche_shape_collapse(self, outputs, show=False, save_path=None):
        durations, _, _, avalanches, bin_size = outputs

        shapes = defaultdict(list)
        for avalanche in avalanches:
            count = avalanche.shape[0]
            if count <= 1:
                continue
            shapes[count].append(avalanche.sum(axis=1))

        fig, axes = plt.subplots(1, 2, figsize=(12, 6))
        for idx, count in enumerate(shapes.keys()):
            shapes[count] = np.array(shapes[count])
            time = np.arange(count) * bin_size
            T = count * bin_size
            mean = np.array(shapes[count]).mean(axis=0)
            err = np.array(shapes[count]).std(axis=0)
            axes[0].errorbar(
                time, mean, yerr=err, label=f"duration {bin_size*count*1000:.2f}ms"
            )
            axes[1].plot(
                time / T,
                mean / (T ** (self.svz - 1)),
                label=f"duration {bin_size*count*1000:.2f}ms",
            )

        axes[0].set_xlabel("Time in Avalanche (s)")
        axes[0].set_ylabel("Average Number of Firing s(t,T)")
        axes[0].set_title("Raw Shapes")
        # axes[0].legend()
        axes[1].set_xlabel("Time in Avalanche Duration (s / T)")
        axes[1].set_ylabel("Average Number of Firing s(t,T) / T^1/svz-1 ")
        axes[1].set_title(f"Normalized/Collapsed Avalanche Shape {self.svz:.2f}")
        # axes[1].legend()

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "avalanche_shape_collapse.png"))
        if show:
            plt.show()
        plt.close(fig)



__all__ = [
    "_Runnable",
    "_RunnerProtocol",
    "VanillaRunner",
    "SupportMultiprocessing",
    "MultiprocessingRunner",
    "InternallyMultiprocessing",
    "StrictMPIRunner",
]

from typing import Any, Callable, Generator, Optional, Protocol, Union

import multiprocessing
import pathlib
from dataclasses import dataclass


class _RunnerProtocol(Callable, Protocol):
    def __call__(self, func: Callable, inputs: Optional[tuple], **kwargs) -> object:
        ...


class _Runnable(Protocol):
    """
    A protocol for a runner policy.
    """

    @property
    def runner(self) -> _RunnerProtocol:
        ...

    def run(
        self,
        save_path: Union[str, pathlib.Path],
        dry_run: bool,
        cache_dir: Union[str, pathlib.Path],
    ) -> None:
        ...


class VanillaRunner:
    """Default runner without any high-level parallelism."""

    def __call__(self, func, inputs=None, **kwargs):
        if inputs is None:
            output = func()
        else:
            output = func(*inputs)
        return output


class MultiprocessingRunner:
    def __init__(self, np: Optional[int] = None):
        if np is None:
            self._np = multiprocessing.cpu_count()
        else:
            self._np = np

    @property
    def num_proc(self):
        return self._np

    def __call__(self, func, inputs: Generator[Any, None, None] = None, **kwargs):
        if inputs is None:
            raise NotImplementedError(
                "Multiprocessing for operator with no generator input is not supported yet. Please use VanillaRunner for this operator."
            )
        else:
            with multiprocessing.Pool(self.num_proc) as p:
                yield from p.imap(func, inputs)


class StrictMPIRunner:
    def __init__(self):
        from mpi4py import MPI

        self.comm = MPI.COMM_WORLD
        self.root = 0

    def set_comm(self, comm):
        self.comm = comm

    def set_root(self, root: int):
        self.root = root

    def get_rank(self):
        return self.comm.Get_rank()

    def get_size(self):
        return self.comm.Get_size()

    def get_root(self):
        return self.root

    def is_root(self):
        return self.get_rank() == self.root

    def __call__(self, func, inputs=None, **kwargs):
        if inputs is None:
            output = func()
        else:
            output = func(*inputs)
        return output


class SupportMPI(StrictMPIRunner):
    pass


class SupportMultiprocessing:
    pass


class InternallyMultiprocessing(Protocol):
    @property
    def num_proc(self) -> int:
        ...
